import logging
import configparser
from os import name
import pathlib
import json

from typing import Sequence, List, Dict, Optional

from ec2init import Plan, Procedure
from ec2init.plans.plan_utils import HopsPlanUtils, REPORT_ENDPOINT
from ec2init.utils.http import HttpUtils
from ec2init.utils.cloud import get_cloud_provider

from ec2init.plans.init.callback_hopsworks import CallbackHopsworks
from ec2init.plans.init.init_ca import InitHopsCA
from ec2init.plans.init.issue_certificate import IssueCertificate
from ec2init.plans.init.configure_glassfish import ConfigureGlassfish
from ec2init.plans.init.start_enable_consul import StartConsul
from ec2init.plans.init.setup_updater import SetupUpdater
from ec2init.plans.init.start_services import StartServices
from ec2init.plans.init.generate_hops_certs import GenerateHopsworksEndpoint
from ec2init.plans.init.check_s3_permissions import CheckS3Permissions
from ec2init.plans.init.check_azure_container_permissions import CheckAzureContainerPermissions
from ec2init.plans.preinit.pre_inits import LocalhostPreInit, MasterPreInit
from ec2init.plans.init.cleanup_pre_init_check import CleanupPreinitCheck
from ec2init.plans.init.setup_eks import SetupEKS, SetupAKS
from ec2init.plans.init.setup_ecr_docker_images import SetupECRDockerImages, SetupAzureDockerImages
from ec2init.plans.init.load_docker import LoadDocker
from ec2init.plans.init.onlinefs_plan import GenerateOnlinefsApiKey
from ec2init.plans.multinode.generate_service_jwt import MasterGenerateServiceJWT, WorkerGenerateServiceJWT
from ec2init.plans.multinode.register_generate_certs import WorkerRegisterGenerateCerts
from ec2init.plans.preinit.setup_cloudwatch_agent import SetupCloudWatchAgent
from ec2init.plans.preinit.database_init import MasterInitializeDatabase
from ec2init.plans.multinode.configure_oauth import ConfigureOAuth
from ec2init.plans.preinit.configure_dns import CentosConfigureDNS
from ec2init.plans.init.user_init_script import UserInitScript
from ec2init.utils.cloud import _DEFAULT_AZURE_SEARCH_DOMAIN
from ec2init.plans.preinit.grow_data_volume import GrowDataVolume
from ec2init.utils.database import databaseUtils
from ec2init.utils.backup_information import BackupInformation
from ec2init.plans.multinode.generate_elastic_signing_key import GenerateElasticSigningKey

logger = logging.getLogger(__name__)


class BaseInitPlan(Plan):
    def init(self, name: str, config: configparser.ConfigParser, **kwargs: str) -> None:
        self.name = name
        config['general']['issue_lets_encrypt'] = kwargs.get("issue_lets_encrypt", "true")
        self.config = config
        self.gateway = kwargs.get("gateway", "gateway")
        self.uid = kwargs.get("uid", "uid")
        self.instance_id = kwargs.get("instance_id", "instance_id")
        self.api_key = kwargs.get("api_key", "")
        self.user_email = kwargs.get("user_email", "user_email")
        self.user_password: Optional[str] = None
        self.password_file = pathlib.Path(kwargs.get(
            "password_file", "/root/ec2init/password"))
        if self.password_file.is_file():
            with open(self.password_file) as pwdfile:
                self.user_password = pwdfile.read().rstrip('\n')
        self.user_mgm_type = kwargs.get("user_management_type", "")
        self.cloud_provider = kwargs.get("cloud_provider", "")
        self.node_id = get_cloud_provider(
            self.cloud_provider).get_node_id(config)
        kwargs['node_id'] = self.node_id
        self.plan_utils = HopsPlanUtils(config, self.user_password, **kwargs)
        self.bucket = kwargs.get("bucket", "")
        self.region = kwargs.get("region", "")

        self.k8s_cluster_name = kwargs.get("k8s_cluster_name", "")
        self.user_registry_url = kwargs.get("user_registry_url", "")
        self.main_registry_url = kwargs.get("main_registry_url", "")
        self.docker_repos = kwargs.get("docker_repos", "")
        self.python_image_name = self.config.get('initplan', 'python_image_name', fallback='python37')
        self.default_registry_url = self.config.get('initplan', 'default_registry_url', fallback='registry.service.consul:4443')
        self.cluster_domain_prefix = kwargs.get("cluster_domain_prefix", "")
        # backup
        self.backup_info = BackupInformation(config)
        backup_info = self.plan_utils.get_procedure_check_file("backup_info")
        if self.plan_utils.check_file_exists(backup_info):
            self.backup_info.load(pathlib.Path(backup_info))
        self.version = kwargs.get("version", "")
        logger.info("%s - version: %s", self.name, self.version)

        self.azure_storage_account = kwargs.get("azure_storage_account", "")
        self.azure_managed_identity_id = kwargs.get("azure_managed_identity_id", "")
        self.azure_acr_username = kwargs.get("azure_acr_username", "")
        self.azure_acr_password = kwargs.get("azure_acr_password", "")
        self.azure_resource_group = kwargs.get("azure_resource_group", "")
        self.collect_logs = kwargs.get("collect_logs", False)
        # Upgrade
        self.source_version = kwargs.get("source_version", "")
        self.source_volume_id = kwargs.get("source_volume_id", "")
        self.is_upgrade = True if self.source_volume_id and self.source_version else False 
        logger.info("%s - is_upgrade: %s", self.name, self.is_upgrade)
        self.deleted_blocks_retention_days = int(kwargs.get("deleted_blocks_retention_days", "0"))
        self.has_public_ip = True if kwargs.get("has_public_ip", "true").lower() == "true" else False
        logger.info("%s - has_public_ip: %s", self.name, self.has_public_ip)
        self.user_init_script = kwargs.get("user_init_script", "")
        logger.info("%s - user_init_script: %s", self.name, self.user_init_script)
        self.search_domain = kwargs.get("azure_search_domain", _DEFAULT_AZURE_SEARCH_DOMAIN)

    def create(self) -> Sequence[Procedure]:
        plan: List[Procedure] = []

        if SetupCloudWatchAgent.is_configured() and self.collect_logs:
            if not self.plan_utils.has_started_enabled_cloudwatch_agent()\
                and not CentosConfigureDNS.should_configure(self.plan_utils):
                logger.info(f"{self.name} - Adding procedure {SetupCloudWatchAgent.__name__}")
                cwa = SetupCloudWatchAgent(SetupCloudWatchAgent.__name__)
                cwa.friendly_name = "Setting CloudWatch agent"
                cwa.report_callback = self.plan_utils.create_callback_procedure(
                    "starting", stage="Setting CloudWatch agent", instance_type=self.get_instance_type())
                plan.append(cwa)
        return plan

    def _get_instance_type(self) -> Optional[str]:
        if hasattr(self, 'instance_type'):
            return getattr(self, 'instance_type', None)
        return None

    def get_instance_type(self) -> str:
        instance_type = self._get_instance_type()
        if instance_type is None:
            raise Exception("instance_type cannot be None")
        return str(instance_type)

    def report_error(self, error_message: str) -> None:
        method, url = REPORT_ENDPOINT
        url = url.format(gateway=self.gateway, uid=self.uid,
                         instance_id=self.instance_id)
        data = {"instanceState": "error", "error": error_message}
        if self.node_id:
            data['nodeId'] = self.node_id
        instance_type = self._get_instance_type()
        if instance_type:
            data['instanceType'] = instance_type
        logger.debug(
            "%s - Reporting an unrecoverable error to %s", self.name, url)
        logger.debug("%s - Error is %s", self.name, data)
        headers: Dict[str, str] = {}
        if self.api_key:
            headers = {"x-api-key": self.api_key}
        try:
            HttpUtils.maybe_api_call(self.config, url, method, data=data, headers=headers)
            logger.debug("%s - Reported error", self.name)
        except Exception as ex:
            logger.error(
                "%s - Could not report error to Hopsworks Cloud: %s", self.name, ex)

    def add_callback_instance_booted(self, plan: List[Procedure], user_email: Optional[str] = None, user_password: Optional[str] = None,
     user_mgm_type: Optional[str] = None) -> None:
        # Make a callback to Hopsworks Cloud indicating the instance has booted
        callback_hopsworks_check = self.plan_utils.get_procedure_check_file(
            "{}_booted".format(CallbackHopsworks.__name__))
        if not self.plan_utils.check_file_exists(callback_hopsworks_check):
            instance_type = self.get_instance_type()
            logger.info("%s - Adding procedure %s with state <booted>",
                        self.name, CallbackHopsworks.__name__)
            cp = CallbackHopsworks(CallbackHopsworks.__name__, self.config, self.gateway, self.instance_id, self.node_id, instance_type,
                                   "booted", self.uid, REPORT_ENDPOINT, self.has_public_ip, user_email, user_password, check_file=callback_hopsworks_check, api_key=self.api_key, 
                                   cloud_provider=self.cloud_provider, user_mgm_type=user_mgm_type)
            cp.fail_plan = True
            cp.friendly_name = "Report instance has booted"
            plan.append(cp)

    def add_grow_data_volume(self, plan: List[Procedure]) -> None:
        check_file = self.plan_utils.get_procedure_check_file(GrowDataVolume.__name__)
        if not self.plan_utils.check_file_exists(check_file):
            logger.info(f"{self.name} - Adding procedure {GrowDataVolume.__name__}")
            grow_vol = GrowDataVolume(GrowDataVolume.__name__, self.config, self.plan_utils)
            grow_vol.friendly_name = "Growing data volume"
            grow_vol.report_callback = self.plan_utils.create_callback_procedure("starting", stage="Growing data volume",
                instance_type=self.get_instance_type())
            plan.append(grow_vol)

    def add_start_consul(self, plan: List[Procedure]) -> None:
        if not self.plan_utils.has_started_enabled_consul():
            logger.info("%s - Adding procedure %s",
                        self.name, StartConsul.__name__)
            cp = StartConsul(StartConsul.__name__)
            plan.append(cp)

    def add_enable_ec2update(self, plan: List[Procedure]) -> None:
        if not self.plan_utils.has_enabled_ec2update_service():
            logger.info("%s - Adding procedure %s",
                        self.name, SetupUpdater.__name__)
            instance_type = self._get_instance_type()
            if instance_type is None:
                raise Exception(
                    "Could not identify instance type. Is it master or worker?")
            instance_type = str(instance_type)
            up = SetupUpdater(SetupUpdater.__name__, self.config, self.instance_id, self.node_id, instance_type,
                              self.uid, self.gateway, self.api_key, self.cloud_provider, self.has_public_ip)
            up.friendly_name = "Setting up ec2update"
            up.report_callback = self.plan_utils.create_callback_procedure(
                "starting", stage="Setting up instance updater", instance_type=instance_type)
            plan.append(up)

    def add_start_enable_services(self, plan: List[Procedure]) -> None:
        if not StartServices.has_started_enabled_all_services(self.config):
            instance_type = self._get_instance_type()
            if instance_type is None:
                raise Exception("instance_type cannot be None")
            logger.info("%s - Adding procedure %s",
                        self.name, StartServices.__name__)
            ss = StartServices(StartServices.__name__, self.gateway, self.instance_id, self.node_id,
                               instance_type, self.uid, self.config, REPORT_ENDPOINT, self.api_key, self.cloud_provider, self.search_domain)
            ss.friendly_name = "Enable and start services"
            ss.report_callback = self.plan_utils.create_callback_procedure(
                "starting", stage="Starting services", instance_type=instance_type)
            plan.append(ss)

    def add_report_all_started(self, plan: List[Procedure]) -> None:
        instance_type = self.get_instance_type()
        report_all_started = self.plan_utils.get_procedure_check_file(
            "{}__started".format(CallbackHopsworks.__name__))
        if not self.plan_utils.check_file_exists(report_all_started) and StartServices.has_started_enabled_all_services(self.config):
            logger.info("%s - Adding procedure %s with state <started>",
                        self.name, CallbackHopsworks.__name__)
            rs = self.plan_utils.create_callback_procedure(
                "started", check_file=report_all_started, instance_type=instance_type)
            rs.friendly_name = "Report instance has started"
            plan.append(rs)

    def add_cleanup_pre_init_check_file(self, plan: List[Procedure]) -> None:
        #check if any of the files that need to be cleaned up exist
        pre_init_check_files = [
            self.plan_utils.get_procedure_check_file(MasterPreInit.__name__),
            self.plan_utils.get_procedure_check_file(LocalhostPreInit.__name__),
            self.plan_utils.get_procedure_check_file("{}_booted".format(CallbackHopsworks.__name__)),
            self.plan_utils.get_procedure_check_file(MasterInitializeDatabase.__name__),
            self.plan_utils.get_procedure_check_file(CentosConfigureDNS.__name__)
        ]

        if self._has_check_file(pre_init_check_files):
            logger.info("%s - Adding procedure %s ", self.name, CleanupPreinitCheck.__name__)
            rs = CleanupPreinitCheck(CleanupPreinitCheck.__name__, pre_init_check_files)
            rs.friendly_name = "Cleanup check files"
            plan.append(rs)

    def _has_check_file(self, check_files: List[pathlib.Path]) -> bool:
        for file in check_files:
            if self.plan_utils.check_file_exists(file):
                return True
        return False

    def add_setup_ecr_docker_images(self, plan: List[Procedure]) -> None:
        if self.user_registry_url and self.main_registry_url and self.docker_repos:
            setup_docker_image_check = self.plan_utils.get_procedure_check_file(SetupECRDockerImages.__name__)
            if not self.plan_utils.check_file_exists(setup_docker_image_check) and not SetupECRDockerImages.is_already_running(SetupECRDockerImages.__name__):
                logger.info("%s - Adding procedure %s", self.name, SetupECRDockerImages.__name__)
                pd = SetupECRDockerImages(SetupECRDockerImages.__name__, check_file=setup_docker_image_check, instance_id=self.instance_id, region=self.region,
                                       ecr_user_registry_url=self.user_registry_url, ecr_main_registry_url=self.main_registry_url, ecr_repos=self.docker_repos,
                                        hopsworks_version=self.version, node_type=self.get_instance_type(), eks_cluster_enabled=bool(self.k8s_cluster_name), 
                                        backup_info=self.backup_info, default_registry_url=self.default_registry_url,
                                        python_image_name=self.python_image_name)
                pd.friendly_name = "Preparing Docker images for ECR"
                pd.report_callback = self.plan_utils.create_callback_procedure("starting", stage="Preparing Docker images for ECR")
                plan.append(pd)
        else:
            logger.info("Skipping ECR setup. No ECR Repo url specified")
    
    def add_setup_azure_docker_images(self, plan: List[Procedure]) -> None:
        if self.user_registry_url and self.main_registry_url and self.docker_repos:
            setup_docker_image_check = self.plan_utils.get_procedure_check_file(SetupAzureDockerImages.__name__)
            if not self.plan_utils.check_file_exists(setup_docker_image_check) and not SetupAzureDockerImages.is_already_running(SetupAzureDockerImages.__name__):
                logger.info("%s - Adding procedure %s", self.name, SetupAzureDockerImages.__name__)
                pd = SetupAzureDockerImages(SetupAzureDockerImages.__name__, check_file=setup_docker_image_check, instance_id=self.instance_id, region=self.region,
                                       ecr_user_registry_url=self.user_registry_url, ecr_main_registry_url=self.main_registry_url, ecr_repos=self.docker_repos,
                                        hopsworks_version=self.version, node_type=self.get_instance_type(), eks_cluster_enabled=bool(self.k8s_cluster_name), 
                                        backup_info=self.backup_info, default_registry_url=self.default_registry_url,
                                        python_image_name=self.python_image_name, azure_acr_username=self.azure_acr_username,
                                        azure_acr_password=self.azure_acr_password, azure_managed_identity_id=self.azure_managed_identity_id)
                pd.friendly_name = "Preparing Docker images for ACR"
                pd.report_callback = self.plan_utils.create_callback_procedure("starting", stage="Preparing Docker images for ACR")
                plan.append(pd)
        else:
            logger.info("Skipping ACR setup. No ACR Repo url specified")

    def add_load_docker_image(self, plan: List[Procedure]) -> None:
        if not (self.user_registry_url and self.main_registry_url):
            load_docker_check = self.plan_utils.get_procedure_check_file(LoadDocker.__name__)
            if not self.plan_utils.check_file_exists(load_docker_check):
                logger.info("%s - Adding procedure %s", self.name, LoadDocker.__name__)
                pks = LoadDocker(LoadDocker.__name__, check_file=load_docker_check, default_registry_url=self.default_registry_url, python_image_name=self.python_image_name,
                 hopsworks_version=self.version)
                pks.friendly_name = "Loading docker image"
                pks.report_callback = self.plan_utils.create_callback_procedure("starting", stage="Loading Docker images")
                plan.append(pks)

    def add_generate_service_jwt(self, plan: List[Procedure]) -> pathlib.Path:
        gen_check = self.plan_utils.get_procedure_check_file(
            WorkerGenerateServiceJWT.__name__)
        if not self.plan_utils.check_file_exists(gen_check)\
            and self.plan_utils.has_started_enabled_consul():
            logger.info("%s - Adding procedure %s", self.name, WorkerGenerateServiceJWT.__name__)
            jwtp = WorkerGenerateServiceJWT(WorkerGenerateServiceJWT.__name__, self.config, gen_check, self.is_upgrade, self.search_domain, self.node_id)
            jwtp.friendly_name = "Generating service tokens"
            jwtp.report_callback = self.plan_utils.create_callback_procedure(
                "starting", stage="Generating service tokens", instance_type=self.get_instance_type())
            plan.append(jwtp)
        return gen_check

    def add_generate_onlinefs_api_key(self, plan: List[Procedure]) -> pathlib.Path:
        gen_check = self.plan_utils.get_procedure_check_file(GenerateOnlinefsApiKey.__name__)
        if not self.plan_utils.check_file_exists(gen_check)\
            and self.plan_utils.has_configured_glassfish():
            logger.info("%s - Adding procedure %s", self.name, GenerateOnlinefsApiKey.__name__)
            gofs = GenerateOnlinefsApiKey(GenerateOnlinefsApiKey.__name__, self.config, self.api_key, gen_check)
            gofs.friendly_name = "Generating Online FS API key"
            gofs.report_callback = self.plan_utils.create_callback_procedure("starting",\
                stage="Generating online FS API key", instance_type=self.get_instance_type())
            plan.append(gofs)
        return gen_check   

    def add_centos_configure_dns(self, plan: List[Procedure]):
        if CentosConfigureDNS.should_configure(self.plan_utils):
            logger.info(f"{self.name} - Adding procedure {CentosConfigureDNS.__name__}")
            ccd = CentosConfigureDNS(CentosConfigureDNS.__name__, self.config, self.plan_utils, self.cloud_provider, self.backup_info)
            ccd.friendly_name = "Configuring DNS"
            ccd.fail_plan = True
            plan.append(ccd)

    def add_user_init_script(self, plan: List[Procedure], call_back: bool=True) -> None:
        # run user init scrip
        user_init_check = self.plan_utils.get_procedure_check_file(UserInitScript.__name__)
        if not self.plan_utils.check_file_exists(user_init_check):
            logger.info("%s - Adding procedure %s", self.name, UserInitScript.__name__)
            userInitPlan = UserInitScript(UserInitScript.__name__, self.config, check_file=user_init_check, user_init_script=self.user_init_script)
            userInitPlan.friendly_name = "User Init Script"
            userInitPlan.fail_plan = True
            if not call_back:
                userInitPlan.report_callback = self.plan_utils.create_callback_procedure("starting", stage="User Init Script", instance_type=self.get_instance_type())
            plan.append(userInitPlan)

class MasterBaseInitPlan(BaseInitPlan):
    def init(self, name: str, config: configparser.ConfigParser, **kwargs: str) -> None:
        super().init(name, config, **kwargs)
        self.ldap_conf_raw = kwargs.get("ldap_conf", "")
        self.ldap_conf = self.plan_utils.base64_to_json(self.ldap_conf_raw)
        self.oauth_conf_raw = kwargs.get("oauth_conf", "")
        self.oauth_conf = self.plan_utils.base64_to_json(self.oauth_conf_raw)
        self.instance_type = "master"
        self.mgmd_ip = kwargs.get("ndb_mgmd_ip", "")

    def create(self) -> Sequence[Procedure]:
        plan : List[Procedure] = []
        parent_plan: Sequence[Procedure] = super().create()
        plan += list(parent_plan)
        
        self.add_grow_data_volume(plan)
        
        return plan

    # user_password argument is ignored but it has to be there to match method signature
    def add_callback_instance_booted(self, plan: List[Procedure], user_email: Optional[str] = None, user_password: Optional[str] = None,
     user_mgm_type: Optional[str] = None) -> None:
        super().add_callback_instance_booted(
            plan, user_email=self.user_email, user_password=self.user_password, user_mgm_type=self.user_mgm_type)


    def add_init_ca(self, plan: List[Procedure]) -> None:
        if not self.plan_utils.has_initialized_hops_ca():
            logger.info("%s - Adding procedure %s",
                        self.name, InitHopsCA.__name__)
            icap = InitHopsCA(InitHopsCA.__name__, self.config)
            icap.friendly_name = "Initializing Hops CA"
            icap.report_callback = self.plan_utils.create_callback_procedure(
                "starting", stage="Initializing Hops CA", instance_type=self.instance_type)
            plan.append(icap)

    def add_issue_LE_certificate(self, plan: List[Procedure]) -> None:
        if not self.plan_utils.has_issued_certificate():
            logger.info("%s - Adding procedure %s",
                        self.name, IssueCertificate.__name__)
            ilec = IssueCertificate(
                IssueCertificate.__name__, self.config, self.instance_id, self.cloud_provider, self.cluster_domain_prefix)
            ilec.friendly_name = "Issuing Let's Encrypt certificate"
            ilec.report_callback = self.plan_utils.create_callback_procedure(
                "starting", stage="Issuing Let's Encrypt certificate", instance_type=self.instance_type)
            plan.append(ilec)

    def add_generate_service_jwt(self, plan: List[Procedure]) -> pathlib.Path:
        gen_check = self.plan_utils.get_procedure_check_file(
            MasterGenerateServiceJWT.__name__)
        if not self.plan_utils.check_file_exists(gen_check) \
            and self.plan_utils.has_configured_glassfish() \
            and self.plan_utils.has_started_enabled_consul():
            logger.info("%s - Adding procedure %s", self.name, MasterGenerateServiceJWT.__name__)
            jwtp = MasterGenerateServiceJWT(MasterGenerateServiceJWT.__name__, self.config, gen_check, BackupInformation.is_backup(self.backup_info) or self.is_upgrade, self.search_domain, self.node_id)
            jwtp.friendly_name = "Generating service tokens"
            jwtp.report_callback = self.plan_utils.create_callback_procedure(
                "starting", stage="Generating service tokens", instance_type=self.instance_type)
            plan.append(jwtp)
        return gen_check

    def add_configure_glassfish(self, plan: List[Procedure]) -> None:
        gen_check = self.plan_utils.get_procedure_check_file(ConfigureGlassfish.__name__)
        required_check_1 = self.plan_utils.get_procedure_check_file(GenerateHopsworksEndpoint.__name__)
        has_configured_glassfish = self.plan_utils.has_configured_glassfish()
        if not self.plan_utils.check_file_exists(gen_check) \
            and self.plan_utils.check_file_exists(required_check_1):
            logger.info("%s - Adding procedure %s", self.name, ConfigureGlassfish.__name__)
            cg = ConfigureGlassfish(ConfigureGlassfish.__name__, self.config, self.ldap_conf, gen_check, has_configured_glassfish)
            cg.friendly_name = "Configuring application server"
            cg.report_callback = self.plan_utils.create_callback_procedure(
                "starting", stage="Configuring application server", instance_type=self.instance_type)
            plan.append(cg)
        
    def add_generate_hopsworks_internal_x509(self, plan: List[Procedure]) -> None:
        gen_check = self.plan_utils.get_procedure_check_file(
            GenerateHopsworksEndpoint.__name__)
        if not self.plan_utils.check_file_exists(gen_check) \
                and self.plan_utils.has_initialized_hops_ca():
            logger.info("%s - Adding procedure %s", self.name,
                        GenerateHopsworksEndpoint.__name__)
            gen_task = GenerateHopsworksEndpoint(
                GenerateHopsworksEndpoint.__name__, self.config, gen_check)
            gen_task.friendly_name = "Generating Hopsworks internal endpoint X.509"
            plan.append(gen_task)

    def add_s3_check(self, plan: List[Procedure]) -> None:
        # Check S3 bucket permissions
        if self.bucket:
            s3_check = self.plan_utils.get_procedure_check_file(
                CheckS3Permissions.__name__)
            s3_test_obj = self.plan_utils.get_cloud_tmp_test_file()

            if not self.plan_utils.check_file_exists(s3_check):
                logger.info("%s - Adding procedure %s", self.name,
                            CheckS3Permissions.__name__)
                s3p = CheckS3Permissions(CheckS3Permissions.__name__, check_file=s3_check,
                                        bucket=self.bucket, region=self.region, s3_test_obj=s3_test_obj, is_backup_or_upgrade=self.is_upgrade or BackupInformation.is_backup(self.backup_info),
                                        deleted_blocks_retention_days=self.deleted_blocks_retention_days)
                s3p.friendly_name = "Checking S3 bucket is empty and the instance has read,write,list and delete privileges to the bucket"
                s3p.fail_plan = True
                s3p.report_callback = self.plan_utils.create_callback_procedure(
                    "starting", stage="Checking S3 permissions", instance_type=self.instance_type)
                plan.append(s3p)
        else:
            logger.info("Skipping S3 checks. No S3 bucket specified")


    def add_azure_container_checks(self, plan: List[Procedure]) -> None:

        # Test Azure container connection string and Permissions
        if self.bucket:
            container_check = self.plan_utils.get_procedure_check_file(
                CheckAzureContainerPermissions.__name__)
            azure_test_obj = self.plan_utils.get_cloud_tmp_test_file()

            if not self.plan_utils.check_file_exists(container_check):
                logger.info("%s - Adding procedure %s", self.name,
                            CheckAzureContainerPermissions.__name__)
                p = CheckAzureContainerPermissions(CheckAzureContainerPermissions.__name__, check_file=container_check,
                                        container=self.bucket, region=self.region, test_obj=azure_test_obj,
                                        storage_account=self.azure_storage_account, client_id=self.azure_managed_identity_id,
                                        deleted_blocks_retention_days=self.deleted_blocks_retention_days, is_backup_or_upgrade=self.is_upgrade or BackupInformation.is_backup(self.backup_info))
                p.friendly_name = "Checking Azure Blob Storage permissions"
                p.fail_plan = True
                p.report_callback = self.plan_utils.create_callback_procedure(
                    "starting", stage="Checking Azure Blob Storage permissions", instance_type=self.instance_type)
                plan.append(p)
        else:
            logger.info("Skipping Azure container checks. No container specified")

    def add_setup_eks(self, plan: List[Procedure], localhost: bool = False) -> None:
        if self.k8s_cluster_name:
            setup_eks_check = self.plan_utils.get_procedure_check_file(SetupEKS.__name__)
            if not self.plan_utils.check_file_exists(setup_eks_check):
                logger.info("%s - Adding procedure %s", self.name, SetupEKS.__name__)
                pks = SetupEKS(SetupEKS.__name__, check_file=setup_eks_check, region=self.region, eks_cluster_name=self.k8s_cluster_name,
                 config=self.config, backup_info=self.backup_info, localhost=localhost)
                pks.friendly_name = "Preparing configurations for EKS"
                pks.fail_plan = True
                pks.report_callback = self.plan_utils.create_callback_procedure("starting", stage="Preparing configurations for EKS")
                plan.append(pks)

    def add_setup_aks(self, plan: List[Procedure], localhost: bool = False) -> None:
        if self.k8s_cluster_name:
            setup_aks_check = self.plan_utils.get_procedure_check_file(SetupAKS.__name__)
            if not self.plan_utils.check_file_exists(setup_aks_check):
                logger.info("%s - Adding procedure %s", self.name, SetupAKS.__name__)
                pks = SetupAKS(SetupAKS.__name__, check_file=setup_aks_check, region=self.region, eks_cluster_name=self.k8s_cluster_name,
                 config=self.config, backup_info=self.backup_info, azure_resource_group=self.azure_resource_group,
                 azure_managed_identity_id=self.azure_managed_identity_id, localhost=localhost)
                pks.friendly_name = "Preparing configurations for AKS"
                pks.fail_plan = True
                pks.report_callback = self.plan_utils.create_callback_procedure("starting", stage="Preparing configurations for AKS")
                plan.append(pks)

    def add_configure_oauth(self, plan: List[Procedure]) -> None:
        if self.oauth_conf and not self.is_upgrade and not BackupInformation.is_backup(self.backup_info):
            gen_check = self.plan_utils.get_procedure_check_file(ConfigureOAuth.__name__)
            if not self.plan_utils.check_file_exists(gen_check) \
                and self.plan_utils.has_configured_glassfish() \
                and self.plan_utils.has_started_enabled_consul():
                logger.info("%s - Adding procedure %s", self.name, ConfigureOAuth.__name__)
                
                oauth = ConfigureOAuth(ConfigureOAuth.__name__, self.config, gen_check, self.oauth_conf, self.has_public_ip,  databaseUtils.ADMIN_EMAIL,
                self.user_password, self.instance_id, self.cloud_provider, self.cluster_domain_prefix)
                oauth.friendly_name = "Configuring OAuth client"
                oauth.report_callback = self.plan_utils.create_callback_procedure("starting", stage="Configuring OAuth client", instance_type=self.instance_type)
                plan.append(oauth)

    def add_generate_elastic_signing_key(self, plan: List[Procedure]) -> pathlib.Path:
        gen_check = self.plan_utils.get_procedure_check_file(
            GenerateElasticSigningKey.__name__)
        if not self.plan_utils.check_file_exists(gen_check) \
            and self.plan_utils.has_configured_glassfish() \
            and self.plan_utils.has_started_enabled_consul():
            logger.info("%s - Adding procedure %s", self.name, GenerateElasticSigningKey.__name__)
            jwtp = GenerateElasticSigningKey(GenerateElasticSigningKey.__name__, self.config, gen_check)
            jwtp.friendly_name = "Generating elastic signing key"
            jwtp.report_callback = self.plan_utils.create_callback_procedure(
                "starting", stage="Generating elastic signing key", instance_type=self.instance_type)
            plan.append(jwtp)
        return gen_check
