import logging
from azure.identity import DefaultAzureCredential
from azure.identity import ManagedIdentityCredential 
import xml.etree.ElementTree as ET
from azure.storage.blob import  BlobServiceClient, BlobClient, ContainerClient, RetentionPolicy
from azure.core.exceptions import ResourceExistsError, ServiceRequestError, HttpResponseError

from typing import List

from ec2init import Procedure
from ec2init.utils.hopsfs import HopsFSUtils
from pathlib import Path

logger = logging.getLogger(__name__)

def fix_hdfs_site_for_azure(storage_account: str, container:str, client_id:str) -> None:
    HopsFSUtils.fix_hdfs_site("dfs.replication", "1")
    HopsFSUtils.fix_hdfs_site("dfs.enable.cloud.persistence", "true")
    HopsFSUtils.fix_hdfs_site("dfs.cloud.provider", "azure")
    HopsFSUtils.fix_hdfs_site("dfs.cloud.azure.container", container)
    HopsFSUtils.fix_hdfs_site("dfs.cloud.azure.storage", storage_account)
    HopsFSUtils.fix_hdfs_site("dfs.disable.non.cloud.storage.policies", "true"); 
    if client_id:
        HopsFSUtils.fix_hdfs_site("dfs.azure.mgm.identity.client.id", client_id)

class CheckAzureContainerPermissions(Procedure):
    def __init__(self, name: str, check_file: Path, container: str, region: str, test_obj: Path, 
                 storage_account: str, client_id:str, 
                 deleted_blocks_retention_days: int, is_backup_or_upgrade: bool) -> None:
        self.name = name
        self.check_file = check_file
        self.container = container 
        self.test_obj = test_obj
        self.region = region
        self.storage_account = storage_account
        self.client_id = client_id
        self.deleted_blocks_retention_days=deleted_blocks_retention_days
        self.is_backup_or_upgrade = is_backup_or_upgrade

    def pre(self) -> None:
        pass

    def execute(self) -> List[Procedure]:
        logger.info("Checking if this instance has proper permissions to access Azure Container: https://%s.blob.core.windows.net/%s",
                    self.storage_account, self.container)
        try :
            # Connect using managed identities
            self.friendly_name = "Unable to get Credentials for Azure"
            if self.client_id:
                logger.info("Using managed identity credentials with client ID: %s", self.client_id)
                credential = ManagedIdentityCredential(client_id=self.client_id)
            else:
                credential = DefaultAzureCredential()

            # Test connection to blob store
            self.friendly_name = "Unable to connect to Azure Blob Storage"
            container_uri = f"https://{self.storage_account}.blob.core.windows.net"
            blob_service_client = BlobServiceClient(account_url=container_uri, credential=credential)


            # Test container exists or create one
            self.friendly_name = "Unable to create container to store data"
            try:
                blob_service_client.create_container(self.container)
            except ResourceExistsError as ex:
                logger.info("Container already exists.")

            # Check Soft Delete support
            self.friendly_name = ("Unable to check support for soft deletes for the storage account. Make sure that your User Assigned Managed Identity has the following access rights on the storage " +
                                  "(Microsoft.Storage/storageAccounts/blobServices/read)")
            self.checkSoftDeletes(blob_service_client)

            # Enable soft deletes 
            if self.deleted_blocks_retention_days > 0:
                self.friendly_name = ("Unable to set backup retention period. Make sure that your User Assigned Managed Identity has the following access rights on the storage or " +
                                      "set the cluster backups validity to 0 (Microsoft.Storage/storageAccounts/blobServices/write, Microsoft.Storage/storageAccounts/blobServices/read)")
                self.enableSoftDeletes(blob_service_client)

            # Test listing. Make sure that container is empty
            self.friendly_name = "Unable to perform list operation on the container"
            container_client = blob_service_client.get_container_client(self.container)
            blobs = container_client.list_blobs()
            count = 0
            for _ in blobs:
                count = count + 1
                break
            
            if not self.is_backup_or_upgrade:
                self.friendly_name = "Unable to start HopsFS as the Azure blob store container is not empty."
                if count > 0 :
                    raise Exception(self.friendly_name)

            # Test uploading to the blob store
            self.friendly_name = "Unable to upload test object to the blob store"
            blob_client = blob_service_client.get_blob_client(container=self.container, blob="test-obj")
            logger.info("Uploading test obj to Azure Storage ")
            with open(self.test_obj, "rb") as data:
                blob_client.upload_blob(data)

            # Test downloading from the blob store
            self.friendly_name = "Unable to download test object from the blob store"
            download_file_path = str(self.test_obj) + ".new" 
            logger.info("Downloading blob to %s" + download_file_path)
            with open(download_file_path, "wb") as download_file:
                download_file.write(blob_client.download_blob().readall())

            # Test deleting an object
            self.friendly_name = "Unable to delete test object from the blob store"
            blob_client = blob_service_client.get_blob_client(container=self.container, blob="test-obj")
            blob_client.delete_blob()
            
            #add azure related properties to hdfs-site.xml
            self.friendly_name = "Unable to set Azure related properties in hdfs-site.xml"
            fix_hdfs_site_for_azure(self.storage_account, self.container, self.client_id)

            #Every thing is OK. 
            if self.check_file is not None:
                self.check_file.touch()

        except HttpResponseError as ex:
            if ex.status_code == 403:
                self.raise_container_exception(f"The managed identity provided does not have permission to access the storage account {self.storage_account}, verify the provided managed identity has the correct permissions.\n"+
                    "For information how to setup the storage account, see https://docs.hopsworks.ai/latest/hopsworksai/azure/getting_started/#step-2-creating-and-configuring-a-storage", ex)
            self.raise_container_exception(str(ex), ex)
        except ServiceRequestError as ex:
            if "Name or service not known" in ex.message:
                self.raise_container_exception(f"The storage account {self.storage_account} could not be found, verify {self.storage_account} is identical to the storage account you created.\n"+
                    "For information how to setup the storage account, see https://docs.hopsworks.ai/latest/hopsworksai/azure/getting_started/#step-2-creating-and-configuring-a-storage", ex)
            self.raise_container_exception(str(ex), ex)
        except Exception as ex:
            self.raise_container_exception(str(ex), ex)

        return []

    def raise_container_exception(self, reason: str, ex: Exception) -> None:
        logger.warn("%s - Could not execute Azure container checks. Container: %s %s", self.name, self.container, ex)
        self.friendly_name = str(self.friendly_name) + "  Reason: " + reason
        raise Exception(self.friendly_name)

    def rollback(self) -> None:
        pass

    def checkSoftDeletes(self, blob_service_client: BlobServiceClient) -> None:
        logger.info("Reading soft delete retention days")
        blob_service_client.get_service_properties()["delete_retention_policy"]

    def enableSoftDeletes(self, blob_service_client: BlobServiceClient) -> None:
        logger.info("Getting soft delete retention days")
        retention_policy = blob_service_client.get_service_properties()["delete_retention_policy"]
        retention_policy_days = 0 if retention_policy.days is None else int(retention_policy.days)
        if retention_policy.enabled and retention_policy_days > self.deleted_blocks_retention_days : 
            raise Exception(f"Retention days for soft deletes are already set to {retention_policy_days}. Lowering the number of retention days for soft deletes will negatively affect existing clusters")

        logger.info("Setting soft delete retention days to %s", str(self.deleted_blocks_retention_days))
        delete_retention_policy = RetentionPolicy(enabled=True, days=self.deleted_blocks_retention_days)
        blob_service_client.set_service_properties(delete_retention_policy=delete_retention_policy)

class CheckAzureContainerWorker(Procedure):
    def __init__(self, name: str, check_file: Path, container: str, storage_account: str, client_id:str) -> None:
        self.name = name
        self.check_file = check_file
        self.container = container 
        self.storage_account = storage_account
        self.client_id = client_id

    def pre(self) -> None:
        pass

    def execute(self) -> List[Procedure]:
        try :
            #add azure related properties to hdfs-site.xml
            self.friendly_name = "Unable to set Azure related properties in hdfs-site.xml"
            fix_hdfs_site_for_azure(self.storage_account, self.container, self.client_id)

            #Every thing is OK. 
            if self.check_file is not None:
                self.check_file.touch()
        except Exception as ex:
            logger.warn("%s - Could not update hdfs-site params for azure. Container: %s %s", self.name, self.container, ex)
            raise Exception(self.friendly_name) 

        return []

    def rollback(self) -> None:
        pass


