import logging
import boto3
import filecmp


from typing import List
from mypy_boto3_s3 import Client

from ec2init import Procedure
from ec2init.utils.system import SystemUtils
from pathlib import Path

logger = logging.getLogger(__name__)

class CheckS3Permissions(Procedure):
    def __init__(self, name: str, check_file: Path, bucket: str, region: str, s3_test_obj: Path, is_backup_or_upgrade: bool,
                deleted_blocks_retention_days: int) -> None:
        self.name = name
        self.check_file = check_file
        self.bucket =  bucket
        self.s3_test_obj = s3_test_obj
        self.region = region
        self.is_backup_or_upgrade = is_backup_or_upgrade
        self.deleted_blocks_retention_days = deleted_blocks_retention_days

    def pre(self) -> None:
        if self.report_callback is not None:
            try:
                self.report_callback.execute()
            except Exception as ex:
                logger.warning("%s - Could not execute pre-execute: %s", self.name, ex)

    def execute(self) -> List[Procedure]:
        logger.info("Checking if this instance has proper permissions to access S3 . Retention days "+ str(self.deleted_blocks_retention_days))
        try :
            self.friendly_name = "Checking S3 Persmisions. Unable to connect to the bucket"
            s3: Client = boto3.client('s3')

            # check that the bucket and VMs are in the same region
            self.friendly_name = "Checking S3 Persmisions. Unable to determine bucket region"
            bl = s3.get_bucket_location(Bucket=self.bucket)["LocationConstraint"]
            if not bl:
                # Buckets in Virginia return null Location
                # https://github.com/logicalclocks/hopsworks-cloud/issues/1417
                self.bucket_location = 'us-east-1'
            else:
                self.bucket_location = str(bl)
            if self.bucket_location != self.region:
                self.friendly_name = "Checking S3 Persmisions. The S3 bucket and VM(s) are in different regions."
                raise Exception(self.friendly_name)

            # HopsFS 3.2.0.3 now requires checking support for versioning even if backups are not enabled. 
            self.isVersioningEnabled(s3)

            if self.deleted_blocks_retention_days > 0:
                if not self.isVersioningEnabled(s3) :
                   self.enableVersioning(s3)

                #put block retention policy
                self.putBlockRetentionPolicy(s3)

            #Check listing permission 
            self.friendly_name = "Checking S3 Persmisions. Unable to perform list operation"
            resp = s3.list_objects_v2(Bucket=self.bucket)

            #Check empty bucket
            keys = resp['KeyCount'] > 0
            if keys > 0 and not self.is_backup_or_upgrade:
                self.friendly_name = "Checking S3 Permissions. The bucket is not empty. Bucket: "  + self.bucket
                raise Exception(self.friendly_name)

            #Check write permission
            self.friendly_name = "Checking S3 Permissions. Unable to upload test object"
            key = "test-obj"
            with open(self.s3_test_obj, 'rb') as data:
                s3.upload_fileobj(data, self.bucket, key)

            #Check read permission
            self.friendly_name = "Checking S3 Permissions. Unable to perform read test" 
            newFile = str(self.s3_test_obj) + ".new" 
            s3.download_file(self.bucket, key, newFile)

            self.friendly_name = "Checking S3 Permissions. Test object is corrupt" 
            if not (filecmp.cmp(str(self.s3_test_obj), newFile)):
                raise Exception(self.friendly_name)

            self.friendly_name = "Checking S3 Permissions. Unable to delete test object"
            s3.delete_object(Bucket=self.bucket, Key=key)

            self.friendly_name = "Checking S3 Permissions. Unable to create multipart upload" 
            multipart_obj_key='multipart-upload-test'
            resp_cmu = s3.create_multipart_upload(Bucket=self.bucket, Key=multipart_obj_key)

            self.friendly_name = "Checking S3 Permissions. Multipart upload failed" 
            resp_up = s3.upload_part(
                Body=b'some bytes',
                Bucket=self.bucket,
                Key=multipart_obj_key,
                PartNumber=1,
                UploadId=resp_cmu['UploadId']
            )


            self.friendly_name = "Checking S3 Permissions. Unable to list multipart uploads" 
            resp_lp = s3.list_multipart_uploads(
                Bucket=self.bucket
            )

            self.friendly_name = "Checking S3 Permissions. Unable to complete multipart upload" 
            s3.complete_multipart_upload(
                Bucket=self.bucket,
                Key=multipart_obj_key,
                MultipartUpload={
                    'Parts': [
                        {
                            'ETag': resp_up['ETag'],
                            'PartNumber': 1 
                        },
                    ]
                },
                UploadId=resp_cmu['UploadId']
            )

            s3.delete_object(Bucket=self.bucket, Key=multipart_obj_key)

            resp_cmu = s3.create_multipart_upload(Bucket=self.bucket, Key=multipart_obj_key)

            self.friendly_name = "Checking S3 Permissions. Unable abort multipart upload" 
            s3.abort_multipart_upload(
                Bucket=self.bucket,
                Key=multipart_obj_key,
                UploadId=resp_cmu['UploadId']
            )

            logger.info("All S3 Permission checks PASSED")

            #Every thing is OK. 
            if self.check_file is not None:
                self.check_file.touch()
        except Exception as ex:
            logger.warning("%s - Could not execute S3 checks. Bucket: %s %s", self.name, self.bucket, ex)
            raise Exception(self.friendly_name) 

        return []

    def rollback(self) -> None:
        pass

    def putBlockRetentionPolicy(self, s3: Client) -> None:
        self.friendly_name = "Checking S3 Persmisions. Unable to set block retention policy"
        response = s3.put_bucket_lifecycle_configuration( # type: ignore 
            Bucket=self.bucket,
            LifecycleConfiguration={
              'Rules': [
                {
                  'Expiration': {
                    'ExpiredObjectDeleteMarker': True
                  }, 
                  'ID': 'HopsFSBlocksRetentionPolicy',
                  'Filter': {},
                  'Status': 'Enabled',
                  'NoncurrentVersionExpiration': {
                    'NoncurrentDays': self.deleted_blocks_retention_days 
                  }, 
                },
              ],
            },
        ) 
        
        logger.info("Put block retention policy response "+str(response))
        if response['ResponseMetadata']['HTTPStatusCode'] != 200:
            raise Exception("Unable to set block retention policy to the S3 bucket. Response: "+str(response))


    def enableVersioning(self, s3: Client) -> None:
        self.friendly_name = "Checking S3 Persmisions. Unable to enable  bucket versioning"
        response = s3.put_bucket_versioning( # type: ignore 
            Bucket=self.bucket,
            VersioningConfiguration={
                'Status': 'Enabled',
            },
        )
        logger.info("Enable versioning response "+str(response))
        if response['ResponseMetadata']['HTTPStatusCode'] != 200:
            raise Exception("Unable to set  bucket versioning. Response: "+str(response))

    def isVersioningEnabled(self, s3: Client) -> bool:
        self.friendly_name = "Checking S3 Persmisions. Unable to check bucket versioning configuration. Add S3:GetBucketVersioning permissions to the instance profile."
        response = s3.get_bucket_versioning( 
            Bucket=self.bucket,
        )
        logger.info("Get versioning response "+str(response))

        if response['ResponseMetadata']['HTTPStatusCode'] != 200:  # type: ignore
            raise Exception("Unable to get bucket versioning information. " +
            "Make sure that you have added S3:GetBucketVersioning permissions to the instance profile. Response: "+str(response))
        else: 
            if 'Status' in response and response['Status']  == 'Enabled':
                return True
            else:
                return False

