import logging
import boto3
import docker
import time 
import kubernetes
import os 
import shutil
import json 

from crontab import CronTab
from pathlib import Path
from typing import List
from mypy_boto3_ecr import ECRClient


from ec2init import MultiThreadProcedure
from ec2init.utils.system import SystemUtils
from ec2init.utils.backup_information import BackupInformation
from ec2init.plans.init.load_docker import LoadDocker

logger = logging.getLogger(__name__)

class SetupDockerImages(MultiThreadProcedure):
    def __init__(self, name: str, check_file: Path, instance_id: str, region: str, ecr_user_registry_url: str, ecr_main_registry_url: str,
     ecr_repos: str, hopsworks_version: str, node_type: str, eks_cluster_enabled: bool, backup_info: BackupInformation,
     default_registry_url: str, python_image_name: str) -> None:
        self.name = name
        self.check_file = check_file
        self.instance_id = instance_id
        self.region = region
        self.hopsworks_version = hopsworks_version
        self.python_base_image_tag = f"{python_image_name}_{self.hopsworks_version}"
        self.tasks = [(repo, hopsworks_version) for repo in ecr_repos.split(",")]
        self.tasks.append(("base", self.python_base_image_tag))
        self.max_workers = min(5, len(self.tasks))
        self._start_time = time.time()
        self.ecr_user_registry_url =  ecr_user_registry_url
        self.ecr_main_registry_url = ecr_main_registry_url
        self.node_type = node_type
        self.eks_cluster_enabled = eks_cluster_enabled
        self.backup_info = backup_info
        # If we're executing this code, the user_registry_url is there so it's safe to cast it from Optional
        self.backup_ecr_user_registry_url = str(backup_info.user_registry_url)
        self.default_registry_url = default_registry_url

    def pre(self) -> None:
        curr_registry_url = self.default_registry_url
        if BackupInformation.is_backup(self.backup_info):
            if self.backup_ecr_user_registry_url == self.ecr_user_registry_url:
                return
            else:
                curr_registry_url = self.backup_ecr_user_registry_url                    

        SystemUtils.execute_command(["sh", "-c", "sed -i 's/docker.trusted.registries=" + curr_registry_url + "/docker.trusted.registries=" + self.ecr_user_registry_url + "/g' /srv/hops/hadoop/etc/hadoop/container-executor.cfg"])

        if self.node_type == "master":
            # update DB
            SystemUtils.execute_command(["/srv/hops/mysql-cluster/ndb/scripts/mysql-client.sh", "hopsworks", "-e", "REPLACE INTO variables (id, value) VALUES ('docker_namespace', '"+ self.instance_id +"')"])

            # registr ECR registry to consul
            registry_consul_file = "/srv/hops/consul/consul.d/registry-consul.hcl"
            registry_consul_conf = ['services = [',
                                    '{',
                                    'id = "registry"',
                                    'name = "registry"',
                                    'port = 443',
                                    'address = "' + self.ecr_user_registry_url + '"',
                                    '}',
                                    ']']
            
            if os.path.exists(registry_consul_file):
                os.remove(registry_consul_file)

            with open(registry_consul_file, 'w') as out_file:
                out_file.writelines(registry_consul_conf)

            os.chmod(registry_consul_file, 0o500)
            shutil.chown(registry_consul_file, group="consul", user="consul")

    def task_function(self, task):
        repo_name, image_tag = task

        user_repo_name = self.instance_id + "/" + repo_name 
        t1 = time.time()
        docker_client = docker.from_env(timeout=360)
        
        if BackupInformation.is_backup(self.backup_info) and self.backup_ecr_user_registry_url and self.node_type == "master":
            self.restore_from_backup(docker_client, repo_name, user_repo_name, image_tag)
        else:
            self.pull_and_push_image(docker_client, repo_name, user_repo_name, image_tag)

        elapsed_time = time.time() - t1
        logger.info("%s took %d seconds to setup image (%s:%s)", self.name, elapsed_time, repo_name, image_tag)

    def pull_and_push_image(self, docker_client, main_repo_name: str, user_repo_name: str, image_tag: str):
        logger.info("%s - Pulling image from hopsworks.ai main repo - %s:%s", self.name, main_repo_name, image_tag)

        main_repo_image = self.ecr_main_registry_url + "/" + main_repo_name + ":" + image_tag
        image = docker_client.images.pull(main_repo_image)
        user_repo_image = self.ecr_user_registry_url + "/" + user_repo_name + ":" + image_tag

        image.tag(user_repo_image)
        docker_client.images.remove(main_repo_image)

        if self.node_type == "master":
            self.push_docker_image(docker_client, user_repo_image)
        
        if self.node_type == "master" and self.eks_cluster_enabled and image_tag == self.python_base_image_tag:
            self.eks_prepull_images(user_repo_image)

    def push_docker_image(self, docker_client, image: str):
        for line in docker_client.images.push(image, stream=True, decode=True):
            if 'error' in line:
                raise Exception(line['error'])

    def restore_from_backup(self, docker_client, main_repo_name: str, user_repo_name: str, image_tag: str):
        t1 = time.time()
        user_repo_image_no_tag = self.ecr_user_registry_url + "/" + user_repo_name
        user_repo_image = user_repo_image_no_tag + ":" + image_tag

        # restoring backup to a different ECR registry
        if self.backup_ecr_user_registry_url != self.ecr_user_registry_url:
            logger.info("%s restore from backup to the another user registry (%s:%s)", self.name, user_repo_name, image_tag)
            old_user_repo_image_no_tag = self.backup_ecr_user_registry_url + "/" + user_repo_name
            old_user_repo_image = old_user_repo_image_no_tag + ":" + image_tag

            old_image = docker_client.images.get(old_user_repo_image)
            old_image.tag(user_repo_image)
            docker_client.images.remove(old_user_repo_image)
            logger.info("%s change tag from %s to %s", self.name, old_user_repo_image, user_repo_image)

            if image_tag == self.python_base_image_tag:
                all_old_user_images = docker_client.images.list(name=old_user_repo_image_no_tag)
                for old_image in all_old_user_images:
                    for old_tag in old_image.tags:
                        new_tag = old_tag.replace(old_user_repo_image_no_tag, user_repo_image_no_tag)
                        if old_tag != new_tag:
                            old_image.tag(new_tag)
                            docker_client.images.remove(old_tag)
                            logger.info("%s change tag from %s to %s", self.name, old_tag, new_tag)
 
        if self.node_type == "master":
            if main_repo_name == "base":
                LoadDocker.load_docker_image(docker_client, user_repo_image, detach=False)

            self.push_docker_image(docker_client, user_repo_image)

            if image_tag == self.python_base_image_tag:
                if self.eks_prepull_images: 
                    self.eks_prepull_images(user_repo_image)

                other_user_images = docker_client.images.list(name=user_repo_image_no_tag)
                for image in other_user_images:
                    for tag in image.tags:
                        self.push_docker_image(docker_client, tag)
                        logger.info("%s push image %s", self.name, tag)
                        
        logger.info("%s took %d seconds to restore backup for %s:%s", self.name, (time.time() - t1), user_repo_name, image_tag)

    def eks_prepull_images(self, image):
        t1 = time.time()        
        kubernetes.config.load_kube_config()
        core_v1 = kubernetes.client.CoreV1Api()
        apps_api = kubernetes.client.AppsV1Api()
        name = 'hopsworks-prepull-' + self.instance_id

        pod_manifest = {
            'apiVersion': 'apps/v1',
            'kind': 'DaemonSet',
            'metadata': {
                'name': name
            },
            'spec': {
                'selector': {
                    'matchLabels': {
                        'name': name
                    }
                },
                'template': {
                    'metadata': {
                        'labels': {
                            'name': name
                        }
                    },
                    'spec': {
                        'initContainers': [{
                            'name': 'prepull',
                            'image': image
                        }],
                        'containers': [{
                            'name': 'pause',
                            'image': 'gcr.io/google_containers/pause'
                        }]
                    }
                }
            }
        }

        hopsworks_namespace = 'hopsworks-' + self.instance_id
        hopsworks_namespace_manifest = {
            'apiVersion': 'v1',
            'kind': 'Namespace',
            'metadata': {
                'name': hopsworks_namespace
            }
        }

        try:
            core_v1.create_namespace(hopsworks_namespace_manifest)
            apps_api.create_namespaced_daemon_set(hopsworks_namespace, pod_manifest)
        except kubernetes.client.rest.ApiException as ex:
            raise_exp = True
            if ex.status == 409:
                if ex.body is not None:
                    resp = json.loads(ex.body)
                    if resp['reason'] == 'AlreadyExists':
                        logger.info("%s - Skipping since %s",
                                    self.name, resp['message'])
                        raise_exp = False
            if raise_exp:
                raise ex
        elapsed_time = time.time() - t1
        logger.info("%s took %d seconds to prepull images on EKS", self.name, elapsed_time)

    def task_rollback_function(self, task):
        pass

    def rollback(self) -> None:
        pass

    def completed_callback(self):
        if self.check_file is not None:
            self.check_file.touch()

        elapsed_time = time.time() - self._start_time
        logger.info("%s took %d seconds to finish setup for %d images", self.name, elapsed_time, len(self.tasks))

class SetupECRDockerImages(SetupDockerImages):
    def __init__(self, name: str, check_file: Path, instance_id: str, region: str, ecr_user_registry_url: str, ecr_main_registry_url: str,
     ecr_repos: str, hopsworks_version: str, node_type: str, eks_cluster_enabled: bool, backup_info: BackupInformation,
     default_registry_url: str, python_image_name: str) -> None:

        super().__init__(name, check_file, instance_id, region, ecr_user_registry_url, ecr_main_registry_url,
            ecr_repos, hopsworks_version, node_type, eks_cluster_enabled, backup_info,
            default_registry_url, python_image_name)

    def task_function(self, task):
        repo_name, image_tag = task
        ecr_client: ECRClient = boto3.client("ecr", region_name=self.region)
        
        user_repo_name = self.instance_id + "/" + repo_name 

        if self.node_type == "master":
            self.create_repository(ecr_client, user_repo_name)
        
        super().task_function(task)

    def create_repository(self, ecr_client: ECRClient, user_repo_name: str):
        try:
            ecr_client.create_repository(repositoryName=user_repo_name, tags=[
                {
                    'Key': 'hopsworks-cluster',
                    'Value': self.instance_id
                }])
        except ecr_client.exceptions.RepositoryAlreadyExistsException:
            logger.info("%s - Skipping repository creation (%s) since it already exists", self.name, user_repo_name)
        
        lifecycle_policy = {
            "rules": [
                {
                    "rulePriority": 1,
                    "description": "Expire images older than 7 days",
                    "selection": {
                        "tagStatus": "untagged",
                        "countType": "sinceImagePushed",
                        "countUnit": "days",
                        "countNumber": 7
                    },
                    "action": {
                        "type": "expire"
                    }
                }
            ]
        }

        lifecycle_policy_text = json.dumps(lifecycle_policy)

        add_policy = True
        try:
            lifecycle_policy_response = ecr_client.get_lifecycle_policy(repositoryName=user_repo_name)
            add_policy = lifecycle_policy_response['lifecyclePolicyText'] != lifecycle_policy_text
            logger.info("%s - Lifecycle policy found for repo %s \n %s", self.name, user_repo_name, str(lifecycle_policy_response))
        except ecr_client.exceptions.LifecyclePolicyNotFoundException:
            logger.info("%s - Lifecycle policy not found for repo %s", self.name, user_repo_name)
            add_policy = True
        
        if add_policy:
            logger.info("%s - Attach a new lifecycle policy to repo %s", self.name, user_repo_name)
            ecr_client.put_lifecycle_policy(repositoryName=user_repo_name, lifecyclePolicyText=lifecycle_policy_text)

     

class SetupAzureDockerImages(SetupDockerImages):
    def __init__(self, name: str, check_file: Path, instance_id: str, region: str, ecr_user_registry_url: str, ecr_main_registry_url: str,
     ecr_repos: str, hopsworks_version: str, node_type: str, eks_cluster_enabled: bool, backup_info: BackupInformation,
     default_registry_url: str, python_image_name: str, azure_acr_username: str, azure_acr_password: str, azure_managed_identity_id: str) -> None:
        self.azure_acr_password = azure_acr_password
        self.azure_acr_username = azure_acr_username
        self.azure_managed_identity_id = azure_managed_identity_id
        # Hardcoded because the AKS images start from version 2.1.0 while hopsworks.ai version was 2.0.0
        self.hopsworks_version = hopsworks_version
        super().__init__(name, check_file, instance_id, region, ecr_user_registry_url, ecr_main_registry_url,
            ecr_repos, self.hopsworks_version, node_type, eks_cluster_enabled, backup_info,
            default_registry_url, python_image_name)

    def pre(self): 
        super().pre()
        logger.info("az login --identity --username %s", self.azure_managed_identity_id)
        SystemUtils.execute_command(["sh", "-c", f"az login --identity --username {self.azure_managed_identity_id}"], timeout=40)
        logger.info("az acr login --name %s", self.ecr_user_registry_url)
        SystemUtils.execute_command(["sh", "-c", f"az acr login --name {self.ecr_user_registry_url}"])
        with CronTab(user='root') as cron:
            job = cron.new(command=f"az acr login --name {self.ecr_user_registry_url}")
            job.hour.every(2)

        if self.node_type == "master":
            SystemUtils.execute_command(["sudo", "-H", "-u", "glassfish", "sh", "-c", f"sudo /srv/hops/sbin/dockerImage.sh loginaz {self.ecr_user_registry_url}"])
            with CronTab(user='glassfish') as cron:
                job = cron.new(command=f"sudo /srv/hops/sbin/dockerImage.sh loginaz {self.ecr_user_registry_url}")
                job.hour.every(2)

        # Docker login using the python api does not persist the credentials, hence use the cli
        # to reduce code dup, see https://github.com/docker/docker-py/issues/1061
        logger.info("docker login -u %s %s", self.azure_acr_username, self.ecr_main_registry_url)
        SystemUtils.execute_command(["sh", "-c", f"docker login -u {self.azure_acr_username} -p {self.azure_acr_password} {self.ecr_main_registry_url}"])