import logging

from typing import List

from ec2init import Procedure
from pathlib import Path
from ec2init.utils.system import SystemUtils
from ec2init.utils.cloud import Azure
from configparser import ConfigParser
from ec2init.utils.cloud import get_cloud_provider, Provider
from ec2init.utils.backup_information import BackupInformation
import kubernetes
from kubernetes.utils.quantity import parse_quantity

logger = logging.getLogger(__name__)

MYSQL_CLIENT = "/srv/hops/mysql-cluster/ndb/scripts/mysql-client.sh"

class SetupK8S(Procedure):
    def __init__(self, name: str, check_file: Path, region: str, eks_cluster_name: str, config: ConfigParser, backup_info: BackupInformation, cloud_provider: str, localhost: bool = False) -> None:
        self.name = name
        self.check_file = check_file
        self.region = region
        self.eks_cluster_name =  eks_cluster_name
        self.config = config
        self.localhost = localhost
        self.backup_info = backup_info
        self.cloud_provider = cloud_provider

    def pre(self) -> None:
        pass

    def execute(self) -> List[Procedure]:
        if self.localhost:
            # Dnsmasq fixes for consul
            curr_ip_config = "127.0.0.2"
            local_ip = get_cloud_provider(self.cloud_provider).get_private_ip(self.config)
            new_ip_config = "127.0.0.2," + local_ip

            if BackupInformation.is_backup(self.backup_info) and self.backup_info.ip:
                curr_ip_config = self.backup_info.ip
                new_ip_config = local_ip
            
            SystemUtils.execute_command(["sh", "-c", "sed -i 's/" + curr_ip_config + "/" + new_ip_config + "/g' /etc/dnsmasq.d/default"])
            SystemUtils.execute_command(["systemctl", "restart", "dnsmasq"])

        #Setup
        self.setup_k8s()

        #Every thing is OK. 
        if self.check_file is not None:
            self.check_file.touch()

        return []

    def setup_k8s(self) -> None:
        kubernetes.config.load_kube_config()
        core_v1 = kubernetes.client.CoreV1Api()
        # Clean up old EKS setup
        self.clean_up_old_eks(core_v1)
        # Setup docker max allocation vars
        self.update_max_docker_allocation(core_v1)

    def clean_up_old_eks(self, core_v1) -> None:
        apps_api = kubernetes.client.AppsV1Api()

        # remove old consul coredns configuration 
        coredns_config_map = core_v1.read_namespaced_config_map("coredns", "kube-system")
        consul_index = coredns_config_map.data['Corefile'].find("consul:53")
        if consul_index > 0:
            logger.info("%s - remove old consul configuration %s", self.name, coredns_config_map.data['Corefile'])
            coredns_config_map.data['Corefile'] = coredns_config_map.data['Corefile'][:consul_index]
            core_v1.patch_namespaced_config_map("coredns", "kube-system", coredns_config_map)
        
        # remove old daemon sets
        all_daemon_sets = apps_api.list_daemon_set_for_all_namespaces()
        all_daemon_sets_names = [(e.metadata.name, e.metadata.namespace) for e in all_daemon_sets.items]
        for name, namespace in all_daemon_sets_names:
            if name.startswith('hopsworks-prepull-') and namespace.startswith('hopsworks-'):
                logger.info("%s - deleting old daemonset (%s, %s)", self.name, name, namespace)
                apps_api.delete_namespaced_daemon_set(name, namespace)
                core_v1.delete_namespace(namespace)

    def update_max_docker_allocation(self, core_v1) -> None:
        kube_docker_max_memory_allocation = 8192
        kube_docker_max_cores_allocation = 4
        kube_docker_cores_fraction = 1.0
        kube_docker_max_gpus_allocation = 0
        try:
            nodes = core_v1.list_node().items 
            if(len(nodes) > 0):
                capacityDict = nodes[0].status.capacity
                if 'cpu' in capacityDict:
                    kube_docker_max_cores_allocation = int(capacityDict['cpu'])
                if 'memory' in capacityDict:
                    kube_docker_max_memory_allocation = int(int(parse_quantity(capacityDict['memory'])) / 1024 / 1024)
                if 'nvidia.com/gpu' in capacityDict:
                    kube_docker_max_gpus_allocation = int(capacityDict['nvidia.com/gpu'])
        except Exception as ex:
            logger.warning("%s - Exception %s", self.name, ex)

        # update variables
        SystemUtils.execute_command([MYSQL_CLIENT, "hopsworks", "-e", f"REPLACE INTO variables (id, value, visibility) VALUES ('kube_docker_max_memory_allocation', '{kube_docker_max_memory_allocation}', '1')"])
        SystemUtils.execute_command([MYSQL_CLIENT, "hopsworks", "-e", f"REPLACE INTO variables (id, value, visibility) VALUES ('kube_docker_max_cores_allocation', '{kube_docker_max_cores_allocation}', '1')"])
        SystemUtils.execute_command([MYSQL_CLIENT, "hopsworks", "-e", f"REPLACE INTO variables (id, value, visibility) VALUES ('kube_docker_cores_fraction', '{kube_docker_cores_fraction}', '1')"])
        SystemUtils.execute_command([MYSQL_CLIENT, "hopsworks", "-e", f"REPLACE INTO variables (id, value, visibility) VALUES ('kube_docker_max_gpus_allocation', '{kube_docker_max_gpus_allocation}', '1')"])


    def rollback(self) -> None:
        pass


class SetupEKS(SetupK8S):
    def __init__(self, name: str, check_file: Path, region: str, eks_cluster_name: str, config: ConfigParser, backup_info: BackupInformation, localhost: bool = False) -> None:
        super().__init__(name, check_file, region, eks_cluster_name, config, backup_info, Provider.AWS, localhost=localhost)

    def execute(self) -> List[Procedure]:
        try :
            aws_command = "aws eks --region " + self.region + " update-kubeconfig --name " + self.eks_cluster_name
            SystemUtils.execute_command(["sh", "-c", aws_command])
            SystemUtils.execute_command(["sudo", "-H", "-u", "glassfish", "sh", "-c", aws_command])
            SystemUtils.execute_command([MYSQL_CLIENT, "hopsworks", "-e", "REPLACE INTO variables (id, value) VALUES ('kube_type', 'eks')"])
            super().execute()
        except Exception as ex:
            logger.warning("%s - Exception %s", self.name, ex)
            raise ex

        return []


class SetupAKS(SetupK8S):
    def __init__(self, name: str, check_file: Path, region: str, eks_cluster_name: str, config: ConfigParser, backup_info: BackupInformation, azure_managed_identity_id: str, azure_resource_group: str, localhost: bool = False) -> None:
        self.azure_managed_identity_id = azure_managed_identity_id
        self.azure_resource_group = azure_resource_group
        super().__init__(name, check_file, region, eks_cluster_name, config, backup_info, Provider.AZURE, localhost=localhost)

    def execute(self) -> List[Procedure]:
        try :
            SystemUtils.execute_command(["sh", "-c", f"az login --identity --username {self.azure_managed_identity_id}"], timeout=40)
            logger.info(f"az aks get-credentials --resource-group {self.azure_resource_group} --name {self.eks_cluster_name}")
            azure_command = f"az aks get-credentials --resource-group {self.azure_resource_group} --name {self.eks_cluster_name}"
            SystemUtils.execute_command(["sh", "-c", azure_command], timeout=40)
            SystemUtils.execute_command(["sudo", "-H", "-u", "glassfish", "sh", "-c", f"az login --identity --username {self.azure_managed_identity_id}"])
            SystemUtils.execute_command(["sudo", "-H", "-u", "glassfish", "sh", "-c", azure_command])
            SystemUtils.execute_command([MYSQL_CLIENT, "hopsworks", "-e", "REPLACE INTO variables (id, value) VALUES ('kube_type', 'aks')"])
            SystemUtils.execute_command([MYSQL_CLIENT, "hopsworks", "-e", "REPLACE INTO variables (id, value) VALUES ('kube_keystore_path', '/usr/lib/jvm/java-8-openjdk-amd64/jre/lib/security/cacerts')"])
            SystemUtils.execute_command([MYSQL_CLIENT, "hopsworks", "-e", "REPLACE INTO variables (id, value) VALUES ('kube_keystore_key', 'changeit')"])
            super().execute()
            self.update_coredns_config()
        except Exception as ex:
            logger.warning("%s - Exception %s", self.name, ex)
            raise ex

        return []

    def update_coredns_config(self) -> None:
        kubernetes.config.load_kube_config()
        core_v1 = kubernetes.client.CoreV1Api()

        consul_server = Azure.get_private_ip(self.config)

        assert consul_server is not None
        coredns_config_map = core_v1.read_namespaced_config_map("coredns-custom", "kube-system")
        if coredns_config_map.data is None:
            coredns_config_map.data = dict()
        coredns_config_map.data['consul.server'] = "consul:53 {\n    errors\n    cache 30\n    forward . "+consul_server+"\n}"
        core_v1.patch_namespaced_config_map("coredns-custom", "kube-system", coredns_config_map)
