import configparser
from ec2init.utils.database import databaseUtils
import logging
import xml.etree.ElementTree as ET
import re

from configparser import ConfigParser
from pathlib import Path

from typing import List, Tuple, Dict, cast

from ec2init import Procedure
from ec2init.utils.http import HttpUtils
from ec2init.utils.dns import DnsUtils
from ec2init.plans.preinit.pre_inits import BasePreInit

logger = logging.getLogger(__name__)

class BaseGenerateServiceJWT(Procedure):
    LOGIN_URI = ("POST", "https://hopsworks.glassfish.service.consul:8182/hopsworks-api/api/auth/service")
    SSL_SERVER_XML = "/srv/hops/hadoop/etc/hadoop/ssl-server.xml"
    RENEW_TOKEN_REGEX = re.compile('hops\.jwt\-manager\.renew\-token\-(?P<idx>[0-9]+)')

    MASTER_TOKEN_DB_ID = "service_master_jwt"
    RENEW_TOKEN_DB_ID = "service_renew_token_"

    def __init__(self, name: str, config: ConfigParser, check_file: Path, is_backup_or_upgrade: bool, search_domain: str, node_id: str) -> None:
        self.name = name
        self.config = config
        self.check_file = check_file
        self.is_backup_or_upgrade = is_backup_or_upgrade
        self.search_domain = search_domain
        self.node_id = node_id

    def pre(self) -> None:
        if self.report_callback is not None:
            try:
                self.report_callback.execute()
            except Exception as ex:
                logger.warning(
                    "%s - Could not execute pre-execute: %s", self.name, ex)

    @staticmethod
    def _get_agent_credentials(config: configparser.ConfigParser) -> Tuple[str, str]:
        kagent_conf_dir = config.get('general', 'kagent_conf_dir', fallback="/srv/hops/kagent/etc")
        kagent_conf_file = Path(kagent_conf_dir, 'config.ini')
        kagent_conf = ConfigParser()
        kagent_conf.read(str(kagent_conf_file))
        username = kagent_conf.get('server', 'username', fallback='agent@hops.io')
        password = kagent_conf.get('server', 'password', fallback='')
        return username, password

    def _replace_service_tokens_in_db(self, master_token: str, renew_tokens: List[str]) -> None:
        logger.debug("%s - Replacing JWT service tokens in database", self.name)
        updates: List[List[str]] = []
        # Empty arguments are fine, we only need the functions for updating Hopsworks variables table
        preInitHelper = BasePreInit(BaseGenerateServiceJWT.__name__, self.config, Path(), '', '', '', self.search_domain, self.node_id)

        if self.is_backup_or_upgrade:
            updates.append(databaseUtils.update_variables_db_cmd(BaseGenerateServiceJWT.MASTER_TOKEN_DB_ID,
            master_token))
        else:
            updates.append(databaseUtils.insert_variables_db_cmd(BaseGenerateServiceJWT.MASTER_TOKEN_DB_ID,
            master_token))

        idx = 0
        for token in renew_tokens:
            rt_id = f"{BaseGenerateServiceJWT.RENEW_TOKEN_DB_ID}{idx}"
            cmd = databaseUtils.update_variables_db_cmd(rt_id, token)
            updates.append(cmd)
            idx += 1

        for cmd in updates:
            preInitHelper.sysexec(cmd)
        
        logger.info("%s - Replaced JWT service tokens in database", self.name)

    def _replace_service_tokens_in_hops(self, master_token: str, renew_tokens: List[str]) -> None:
        logging.debug("%s - Replacing JWT service tokens in Hops ssl-server.xml", self.name)
        tree = ET.parse(BaseGenerateServiceJWT.SSL_SERVER_XML)
        root = tree.getroot()
        for p in tree.iter('property'):
            if p[0].text == 'hops.jwt-manager.master-token':
                p[1].text = master_token
            else:
                m = BaseGenerateServiceJWT.RENEW_TOKEN_REGEX.match(str(p[0].text))
                if m:
                    idx = m.group('idx')
                    p[1].text = renew_tokens[int(idx)]

        tree.write(BaseGenerateServiceJWT.SSL_SERVER_XML)
        logging.info("%s - Replaced JWT service tokens in Hops ssl-server.xml", self.name)

    def _get_tokens(self) -> Tuple[str, List[str]]:
        return BaseGenerateServiceJWT.get_service_jwt_tokens(self.config)

    @staticmethod
    def get_service_jwt_tokens(config : ConfigParser) -> Tuple[str, List[str]]:
        agent_username, agent_password = BaseGenerateServiceJWT._get_agent_credentials(config)
        if not agent_username or not agent_password:
            raise Exception("Agent username or password is not set")

        method, uri = BaseGenerateServiceJWT.LOGIN_URI
        data = {
            'email': agent_username,
            'password': agent_password,
        }
        headers = {
            'Content-Type': 'application/x-www-form-urlencoded',
        }

        response: Dict[str, str] = HttpUtils.do_plain_call(config, uri, method, data=data, headers=headers, verify=False)
        response_headers: Dict[str, str] = cast(Dict[str, str], response['headers'])
        # This is form 'Bearer TOKEN'
        master_bearer_token = response_headers['Authorization']
        master_token = master_bearer_token.split(" ")[1].strip()
        content: Dict[str, str] = cast(Dict[str, str], response['content'])
        renew_tokens: List[str] = cast(List[str], content['renewTokens'])

        return master_token, renew_tokens

    def rollback(self):
        pass


class MasterGenerateServiceJWT(BaseGenerateServiceJWT):
    def execute(self) -> List[Procedure]:
        logger.info("%s - Starting execution", self.name)

        DnsUtils.get_ip_from_name("hopsworks.glassfish.service.consul")

        master_token, renew_tokens = self._get_tokens()
        self._replace_service_tokens_in_db(master_token, renew_tokens)

        master_token, renew_tokens = self._get_tokens()
        self._replace_service_tokens_in_hops(master_token, renew_tokens)

        if self.check_file is not None:
            self.check_file.touch()
        logger.info("%s - Finished execution", self.name)
        return []

class WorkerGenerateServiceJWT(BaseGenerateServiceJWT):
    def execute(self) -> List[Procedure]:
        logger.info("%s - Starting execution", self.name)

        DnsUtils.get_ip_from_name("hopsworks.glassfish.service.consul")
        master_token, renew_tokens = self._get_tokens()
        self._replace_service_tokens_in_hops(master_token, renew_tokens)

        if self.check_file is not None:
            self.check_file.touch()
        logger.info("%s - Finished execution", self.name)
        return []