import logging
import pathlib
import subprocess

from configparser import ConfigParser

from typing import Sequence, List

from ec2init import Procedure
from ec2init.plans.init.base_init_plan import MasterBaseInitPlan

from ec2init.plans.multinode.master_seds import MasterSeds
from ec2init.plans.multinode.register_generate_certs import RegisterGenerateCerts
from ec2init.utils import cloud
from ec2init.utils.backup_information import BackupInformation
from ec2init.utils.system import SystemUtils


logger = logging.getLogger(__name__)

class MasterInitPlan(MasterBaseInitPlan):
    def init(self, name: str, config: ConfigParser, **kwargs: str) -> None:
        super().init(name, config, **kwargs)

    def create(self) -> Sequence[Procedure]:
        plan: List[Procedure] = []

        # Replace master node IP placeholder with the actual IP
        sed_check = self.plan_utils.get_procedure_check_file(MasterSeds.__name__)
        if not self.plan_utils.check_file_exists(sed_check):
            logger.info("%s - Adding procedure %s", self.name, MasterSeds.__name__)
            sed_task = MasterSeds(MasterSeds.__name__, self.config, sed_check, self.cloud_provider, self.backup_info,
                self.instance_id, self.search_domain, self.node_id, self.cluster_domain_prefix)
            sed_task.friendly_name = "Updating addresses"
            sed_task.report_callback = self.plan_utils.create_callback_procedure("starting", stage="Updating addresses", instance_type = self.instance_type)
            plan.append(sed_task)

        # In case of backup restore we would like to recover S3 bucket
        if BackupInformation.is_backup(self.backup_info):
            logger.info("Restoring the deleted blocks")
            subprocess.run(["su - hdfs -c '/srv/hops/hadoop/bin/hdfs namenode -rollbackDeletedCloudBlocks'" ], stdout=subprocess.PIPE, shell=True)

        # Initialize Hops CA
        self.add_init_ca(plan)

        # Generate Hopsworks internal endpoint certificate
        self.add_generate_hopsworks_internal_x509(plan)

        # Issue Let's Encrypt certificate and setup
        self.add_issue_LE_certificate(plan)

        # Configure Glassfish to use previously issued certificate and setup hook for auto-renew
        self.add_configure_glassfish(plan)

        self.add_start_consul(plan)

        # Generate Service tokens if glassfish has started
        service_jwt_check = self.add_generate_service_jwt(plan)

        # Generate Elastic signing key 
        self.add_generate_elastic_signing_key(plan)
        
        # Generate Online FS API key
        onlinefs_api_key_check = self.add_generate_onlinefs_api_key(plan)

        # Regenerate Hops host certificate
        register_generate_cert_check = self.plan_utils.get_procedure_check_file(RegisterGenerateCerts.__name__)
        if not self.plan_utils.check_file_exists(register_generate_cert_check) \
            and self.plan_utils.has_initialized_hops_ca() and self.plan_utils.check_file_exists(service_jwt_check):
            logger.info("%s - Adding procedure %s", self.name, RegisterGenerateCerts.__name__)
            p3 = RegisterGenerateCerts(RegisterGenerateCerts.__name__, self.config, register_generate_cert_check,
                                       self.cloud_provider, self.api_key, self.gateway, self.uid, self.instance_id, self.node_id, self.search_domain)
            p3.friendly_name = "Generating Hops x.509"
            p3.report_callback = self.plan_utils.create_callback_procedure("starting", stage="Generating Hops x.509", instance_type = self.instance_type)
            plan.append(p3)

        # Setup ec2update systemd unit to callback Hopsworks Cloud when the instance has restared
        self.add_enable_ec2update(plan)

        # Configure OAuth 
        self.add_configure_oauth(plan)

        # Enable and start all systemd units listed in kagent's services file
        # This should always be the last Procedure, as it signals Hopsworks Cloud
        # that the instance is up and running
        if self._has_ran_all(sed_check, register_generate_cert_check, service_jwt_check, onlinefs_api_key_check):
            self.add_start_enable_services(plan)

        return plan

    ## Add you checks here!
    def _has_ran_all(self, sed_check: pathlib.Path, regenerate_hops_x509_check_file: pathlib.Path, service_jwt_check: pathlib.Path, onlinefs_api_key_check: pathlib.Path) -> bool:
        return self.plan_utils.check_file_exists(sed_check) and self.plan_utils.has_initialized_hops_ca() and self.plan_utils.has_issued_certificate() \
            and self.plan_utils.has_configured_glassfish() and self.plan_utils.has_started_enabled_consul() and self.plan_utils.has_enabled_ec2update_service() \
            and self.plan_utils.check_file_exists(regenerate_hops_x509_check_file) and self.plan_utils.check_file_exists(service_jwt_check) \
            and self.plan_utils.check_file_exists(onlinefs_api_key_check)

class AWSMasterInitPlan(MasterInitPlan):
    def init(self, name: str, config: ConfigParser, **kwargs: str) -> None:
        # The cloud provider attribute may have already been set, but we make sure it is correct here instead
        kwargs['cloud_provider'] = cloud.Provider.AWS
        super().init(name, config, **kwargs)

    def create(self) -> Sequence[Procedure]:
        plan: List[Procedure] = []

        # Check S3 bucket permissions
        self.add_s3_check(plan)

        # Setup EKS Cluster
        self.add_setup_eks(plan)

        # Setup ECR Docker images 
        self.add_setup_ecr_docker_images(plan)

        # Load the docker image from the AMI
        self.add_load_docker_image(plan)

        parentPlan: Sequence[Procedure] = super().create()

        return plan + list(parentPlan)


class AzureMasterInitPlan(MasterInitPlan):
    def init(self, name: str, config: ConfigParser, **kwargs: str) -> None:
        # The cloud provider attribute may have already been set, but we make sure it is correct here instead
        kwargs['cloud_provider'] = cloud.Provider.AZURE
        super().init(name, config, **kwargs)

    def create(self) -> Sequence[Procedure]:
        plan: List[Procedure] = []
        # Check for Azure Block storage permissions
        self.add_azure_container_checks(plan)

        self.add_setup_aks(plan)
        self.add_setup_azure_docker_images(plan)
        self.add_load_docker_image(plan)
        parentPlan: Sequence[Procedure] = super().create()
        return plan + list(parentPlan)

class GCPMasterInitPlan(MasterInitPlan):
    def init(self, name: str, config: ConfigParser, **kwargs: str) -> None:
        # The cloud provider attribute may have already been set, but we make sure it is correct here instead
        kwargs['cloud_provider'] = cloud.Provider.GCP
        super().init(name, config, **kwargs)

    def create(self) -> Sequence[Procedure]:
        plan: List[Procedure] = []
        parentPlan: Sequence[Procedure] = super().create()
        return plan + list(parentPlan)
