import logging
import configparser
import subprocess

from pathlib import Path
from typing import List

from ec2init import Procedure
from ec2init.plans.multinode.seds import Seds
from ec2init.utils.system import SystemUtils
from ec2init.utils.cloud import get_cloud_provider
from ec2init.utils.backup_information import BackupInformation

logger = logging.getLogger(__name__)

class MasterSeds(Seds):
    def __init__(self, name: str, config: configparser.ConfigParser, check_file: Path, cloud_provider: str, backup_info: BackupInformation,
      cluster_id: str, search_domain: str, node_id: str, cluster_domain_prefix: str) -> None:
        super().__init__(name, config, cloud_provider, search_domain, node_id, backup_info)
        self.check_file = check_file
        self.cluster_id = cluster_id
        self.cluster_domain_prefix = cluster_domain_prefix

    def execute(self) -> List[Procedure]:
        logger.info("%s - Starting execution", self.name)
        children_procs: List[Procedure] = []

        super_procedures = super().execute()

        children_procs = super_procedures + children_procs

        public_domain_name = get_cloud_provider(self.cloud_provider)\
            .get_public_domain_name(self.config, self.cluster_id, self.cluster_domain_prefix)

        subString = "s/{0}/{1}/g".format(self.MASTER_IP, self.local_ip)

        commands: List[str] = []

        ndbprops = Seds.REPLACE.format(subString, '/srv/hops/hadoop/etc/hadoop/ndb.props')
        commands.append(ndbprops)

        epipec = Seds.REPLACE.format(subString, '/srv/hops/epipe/conf/config.ini')
        commands.append(epipec)

        epipecr = Seds.REPLACE.format(subString, '/srv/hops/epipe/conf/config-reindex.ini')
        commands.append(epipecr)

        kafkap = Seds.REPLACE.format(subString, '/srv/hops/kafka/config/server.properties')
        commands.append(kafkap)

        kafkaExternalReplacement = f"s/{self.PUBLIC_DOMAIN_NAME}/{public_domain_name}/g"
        kafkaExternal = Seds.REPLACE.format(kafkaExternalReplacement, '/srv/hops/kafka/config/server.properties')
        commands.append(kafkaExternal)

        livyc = Seds.REPLACE.format(subString, '/srv/hops/apache-livy/conf/livy.conf')
        commands.append(livyc)

        hostnameSubString = "s/{0}/{1}/g".format(self.MASTER_HOSTNAME, self.fqdn)
        kafkaph = Seds.REPLACE.format(hostnameSubString, "/srv/hops/kafka/config/server.properties")
        commands.append(kafkaph)

        elastic = Seds.REPLACE.format(hostnameSubString, "/srv/hops/elasticsearch/config/elasticsearch.yml")
        commands.append(elastic)

        runSecurityAdmin = Seds.REPLACE.format(hostnameSubString, "/srv/hops/elasticsearch/plugins/opendistro_security/tools/run_securityAdmin.sh")
        commands.append(runSecurityAdmin)
        
        elastic_exporter = Seds.REPLACE.format(hostnameSubString, "/lib/systemd/system/elastic_exporter.service")
        commands.append(elastic_exporter)
        
        for c in commands:
            logger.debug("%s - Executing command %s", self.name, c)
            SystemUtils.execute_command(c.split(" "))
            logger.info("%s - Executed command %s", self.name, c)

        logstash = Seds.SEEK_AND_REPLACE.format("/srv/hops/logstash/config", hostnameSubString)
        # Unix find is a pain for subprocess
        subprocess.run(logstash, timeout=20, shell=True)

        self.sed_almost_similar(self.local_ip, self.fqdn)
        
        if self.check_file:
            self.check_file.touch()

        return children_procs