import logging
import configparser
import pathlib

from typing import Sequence, List, Type

from ec2init import Procedure
from ec2init.plans.init.base_init_plan import BaseInitPlan
from ec2init.plans.init.start_enable_consul import StartConsul
from ec2init.plans.preinit.pre_inits import BasePreInit
from ec2init.plans.multinode.seds import Seds
from ec2init.plans.multinode.register_generate_certs import NdbNodeRegisterGenerateCerts
from ec2init.plans.ndb.start_side_services import StartSideServices

from ec2init.utils.cloud import get_cloud_provider, CloudProvider
from ec2init.utils.system import OSType, SystemUtils

logger = logging.getLogger(__name__)

class NdbNodeJoinCluster(BaseInitPlan):
    def init(self, name: str, config: configparser.ConfigParser, **kwargs: str) -> None:
        super().init(name, config, **kwargs)
        hip = kwargs.get("master_ip", None)
        if not hip:
            raise Exception("Hopsworks head IP address is missing. Did you pass master_ip argument?")
        self.head_ip = str(hip)

        ## TODO: Not a good idea hardcoding the password
        self.key_password = "adminpw"

    def create(self) -> Sequence[Procedure]:
        plan: List[Procedure] = []
        cloud_provider = get_cloud_provider(self.cloud_provider)

        replacements_check_file = self.plan_utils\
            .get_procedure_check_file(ClusterJoinReplacements.__name__)
        if not self.plan_utils.check_file_exists(replacements_check_file):
            cjr = ClusterJoinReplacements(ClusterJoinReplacements.__name__,
                self.config, replacements_check_file, self.api_key, cloud_provider, self.head_ip, self.search_domain, self.node_id)
            cjr.fail_plan = True
            cjr.report_callback = None
            plan.append(cjr)

        certs_check_file = self.plan_utils.get_procedure_check_file(NdbNodeRegisterGenerateCerts.__name__)
        if self.plan_utils.check_file_exists(replacements_check_file) \
            and not self.plan_utils.check_file_exists(certs_check_file):
                gen_certs = NdbNodeRegisterGenerateCerts(NdbNodeRegisterGenerateCerts.__name__,
                    self.config, certs_check_file, self.head_ip, self.cloud_provider, self.key_password, 
                    self.api_key, self.gateway, self.uid, self.instance_id, self.node_id, self.search_domain)
                gen_certs.report_callback = None
                plan.append(gen_certs)

        if self.plan_utils.check_file_exists(certs_check_file):
            if not self.plan_utils.has_started_enabled_consul():
                sc = StartConsul(StartConsul.__name__)
                sc.report_callback = None
                plan.append(sc)

        if self.plan_utils.check_file_exists(certs_check_file):
            if not StartSideServices.has_started_services():
                ss = StartSideServices(StartSideServices.__name__)
                ss.report_callback = None
                plan.append(ss)

        return plan

class ClusterJoinReplacements(Procedure):
    def __init__(self, name: str, config: configparser.ConfigParser, check_file: pathlib.Path,
        api_key: str, cloud_provider: Type[CloudProvider], head_ip: str, search_domain: str, node_id: str):
        self.config = config
        self.check_file = check_file
        self.api_key = api_key
        self.cloud_provider = cloud_provider
        self.head_ip = head_ip
        self.search_domain = search_domain
        self.node_id = node_id

    def execute(self) -> List[Procedure]:
        ## We only need a specific method, it's fine to pass empty arguments
        preInit = BasePreInit(ClusterJoinReplacements.__name__, self.config, self.check_file,
            "", "", "", self.search_domain, self.node_id)
        preInit.update_kagent_password(self.api_key)

        local_ip = self.cloud_provider.get_private_ip(self.config)
        fqdn = self.cloud_provider.get_fqdn(self.config, self.search_domain, self.node_id)
        seds = Seds(ClusterJoinReplacements.__name__, self.config, "", self.search_domain, self.node_id)
        seds._replaceNodeFqdnConsulBashrc(fqdn)
        seds._replaceNodeFqdnConsulService(fqdn)
        seds._replaceNodeIpDnsmasq(local_ip)
        if SystemUtils.get_os_type() == OSType.UBUNTU:
            seds._replaceNodeIpSystemdResolved(local_ip)

        ## Replace retry join /srv/hops/consul/consul.d/consul.hcl
        replacementStr = f"s;^retry_join[[:space:]]=.*$;retry_join = [\"{self.head_ip}\"];g"
        cmd = [
            "sudo",
            "sed",
            "-i",
            replacementStr,
            "/srv/hops/consul/consul.d/consul.hcl"
        ]
        SystemUtils.execute_command(cmd)

        if self.check_file:
            self.check_file.touch()

        return []

    def rollback(self) -> None:
        pass
    