import logging
import configparser

from typing import List, Optional

from ec2init import Procedure
from ec2init.utils.system import SystemUtils
from ec2init.plans.multinode.seds import Seds
from ec2init.plans.init.start_services import StartServices

logger = logging.getLogger(__name__)

HEARTBEAT_ENDPOINT = "{gateway}/eventsv2/ndbagent/{uid}/{cid}/heartbeat"

class BaseStartNdbAgent(Procedure):
    AGENT_CONFIG="/srv/hops/ndb-agent/ndb-agent/config/config.yml"
    def __init__(self, name: str, config: configparser.ConfigParser, api_key: str, hb_endpoint: Optional[str]) -> None:
        self.name = name
        self.config = config
        self.api_key = api_key
        self.hb_endpoint = hb_endpoint


    def execute(self) -> List[Procedure]:
        logger.debug(f"{self.name} - Replacing API key")
        subString = f"s/API_KEY/{self.api_key}/g"
        cmd = Seds.REPLACE.format(subString, BaseStartNdbAgent.AGENT_CONFIG)
        SystemUtils.execute_command(cmd.split(" "))
        logger.info(f"{self.name} - Replaced API key")

        if self.hb_endpoint:
            logger.debug(f"{self.name} - Enabling Heartbeat")
            subString = f"s/HEARTBEAT_ENABLED/true/g"
            cmd = Seds.REPLACE.format(subString, BaseStartNdbAgent.AGENT_CONFIG)
            SystemUtils.execute_command(cmd.split(" "))
            logger.info(f"{self.name} - Enabled Heartbeat")

            logger.debug(f"{self.name} - Replacing Heartbeat endpoint")
            subString = f"s;HEARTBEAT_ENDPOINT;{str(self.hb_endpoint)};g"
            cmd = Seds.REPLACE.format(subString, BaseStartNdbAgent.AGENT_CONFIG)
            SystemUtils.execute_command(cmd.split(" "))
            logger.info(f"{self.name} - Replaced Heartbeat endpoint")
        else:
            logger.debug(f"{self.name} - Disabling Heartbeat")
            subString = f"s/HEARTBEAT_ENABLED/false/g"
            cmd = Seds.REPLACE.format(subString, BaseStartNdbAgent.AGENT_CONFIG)
            SystemUtils.execute_command(cmd.split(" "))
            logger.info(f"{self.name} - Disabled Heartbeat")

        logger.debug(f"{self.name} - Starting ndb-agent")
        SystemUtils.execute_command(["systemctl", "enable", "ndb-agent"], timeout=120)
        SystemUtils.execute_command(["systemctl", "start", "ndb-agent"], timeout=120)
        logger.info(f"{self.name} - Started ndb-agent")

        return []

    def rollback(self):
        pass

    @staticmethod
    def has_run() -> bool:
        return StartServices._is_service_running("ndb-agent")\
            and StartServices.is_service_enabled("ndb-agent")

class BootstrapMgmStartNdbAgent(BaseStartNdbAgent):
    def __init__(self, name: str, config: configparser.ConfigParser, api_key: str, gateway: str, uid: str, cluster_id: str) -> None:
        hb_endpoint = HEARTBEAT_ENDPOINT.format(gateway=gateway, uid=uid, cid=cluster_id)
        super().__init__(name, config, api_key, hb_endpoint)

class StartNdbAgent(BaseStartNdbAgent):
    def __init__(self, name: str, config: configparser.ConfigParser, api_key: str) -> None:
        super().__init__(name, config, api_key, None)