from ec2init.utils.database import databaseUtils
import logging
import time
from typing import List, Optional
from ec2init import Procedure
from ec2init.utils.http import HttpUtils
from ec2init.utils.system import SystemUtils
from pathlib import Path
from configparser import ConfigParser
from ec2init.utils.system import SystemUtils

logger = logging.getLogger(__name__)

class PostInit(Procedure):
    def __init__(self, name: str, config: ConfigParser, check_file: Path, bucket: str, is_upgrade: bool, user_management_type: str, user_password: Optional[str]) -> None:
        self.name = name
        self.config = config
        self.check_file = check_file
        self.bucket = bucket
        self.is_upgrade = is_upgrade
        self.user_management_type = user_management_type
        self.user_password = user_password
        self.run_securityadmin_script = Path("/srv/hops/elasticsearch/plugins/opendistro_security/tools/run_securityAdmin.sh")
        self.base_docker_images = Path(self.config.get('upgradeplan', 'base_images', fallback='/root/base.tar'))
        self.onlinefs_docker_image = Path(self.config.get('upgradeplan', 'onlinefs_image', fallback='/root/onlinefs.tgz'))

    def pre(self) -> None:
        if self.report_callback is not None:
            try:
                self.report_callback.execute()
            except Exception as ex:
                logger.warning("%s - Could not execute pre-execute: %s", self.name, ex)

    def execute(self) -> List[Procedure]:
        try:
            oauth_enabled = SystemUtils.execute_mysql_statement("SELECT value from variables WHERE id='oauth_enabled'")[0][0].upper() == "TRUE"
            should_disable_password_login = oauth_enabled if self.user_management_type.upper() == "MANAGED" else self.user_management_type.upper() != "DISABLED"
            if should_disable_password_login:
                jwt = databaseUtils._login(databaseUtils.ADMIN_EMAIL, self.user_password, self.config) #can we save the jwt for retry?
                databaseUtils.update_variable("disable_password_login", "true")
                databaseUtils.refresh_variables(self.config, jwt)
            #block default admin account"
            databaseUtils.run_command(["hopsworks", "-e", 
                    "UPDATE users SET status=4 where username='meb10000'"])
                
            #wait until NN is out of safemode
            #/srv/hops/hadoop/bin/hdfs dfsadmin -safemode get
            #       Safe mode is OFF
            #       Safe mode is ON 
            #If the namenode is not started then NN throws IOException with the following message
            #Leader Node still not started
            for _ in range(0, 24): 
                try:
                    process = SystemUtils.execute_hdfs_command("dfsadmin -safemode get")
                    logger.info("Safemode check output: "+process.stdout.strip().decode("utf-8"))
                    output = process.stdout.strip().decode("utf-8")
                    if "Safe mode is OFF" in output:
                        break
                    else:
                        logger.warn("NN is in safemode or it has not yet started. Retrying ...")
                        time.sleep(5)

                except Exception as ex:
                    logger.warning("%s - Namenode is not up.", ex)
                    time.sleep(5)


            if self.bucket and not self.is_upgrade:
                SystemUtils.execute_hdfs_command("mover -p /")
            
            if self.run_securityadmin_script.is_file():
                logger.info("%s - runing elastic securityAdmin", self.name)
                SystemUtils.execute_command([self.run_securityadmin_script.as_posix()], shell=True, timeout=None)

            # Remove temp base.tar
            if self.base_docker_images.is_file():
                self.base_docker_images.unlink()
                logger.info("%s - Deleted the tar file for base docker images %s", self.name, self.base_docker_images)

            if self.onlinefs_docker_image.is_file():
                self.onlinefs_docker_image.unlink()
                logger.info("%s - Deleted the tar file for onlinefs docker images %s", self.name, self.onlinefs_docker_image)

            # Every thing is OK.
            if self.check_file is not None:
                self.check_file.touch()
        except Exception as ex:
            logger.warning("%s - Could not initialize cluster.", ex)
            raise ex

        return []

    def rollback(self) -> None:
        pass
