import logging
from configparser import ConfigParser

from typing import Sequence, List

from ec2init import Procedure
from ec2init.plans.init.base_init_plan import BaseInitPlan
from ec2init.plans.postinit.post_init import PostInit
from ec2init.plans.init.user_init_script import UserInitScript 
from ec2init.plans.postinit.store_backup_info import StoreBackupInfo
from ec2init.plans.postinit.expat import Expat 

logger = logging.getLogger(__name__)


class BasePostInitPlan(BaseInitPlan):
    def init(self, name: str, config: ConfigParser, **kwargs: str) -> None:
        super().init(name, config, **kwargs)

#for master but can't be renamed for backward compatibility reasons
class PostInitPlan(BasePostInitPlan):
    def init(self, name: str, config: ConfigParser, **kwargs: str) -> None:
        super().init(name, config, **kwargs)
        self.instance_type = "master"

    def create(self) -> Sequence[Procedure]:
        plan: List[Procedure] = []

        backup_info_check = self.plan_utils.get_procedure_check_file(StoreBackupInfo.__name__)
        if not self.plan_utils.check_file_exists(backup_info_check):
            backupInfoPlan = StoreBackupInfo(StoreBackupInfo.__name__, self.config, self.plan_utils.get_procedure_check_file(
                "backup_info"), backup_info_check, self.cloud_provider, self.user_registry_url, self.search_domain, self.node_id)
            backupInfoPlan.friendly_name = "Store backup Info"
            backupInfoPlan.fail_plan = True
            backupInfoPlan.report_callback = self.plan_utils.create_callback_procedure("starting", stage="Store backup Info", instance_type=self.get_instance_type())
            plan.append(backupInfoPlan)

        post_init_check = self.plan_utils.get_procedure_check_file(PostInit.__name__)
        if not self.plan_utils.check_file_exists(post_init_check):
            logger.info("%s - Adding procedure %s", self.name, PostInit.__name__)
            postInitPlan = PostInit(PostInit.__name__, self.config, check_file=post_init_check, bucket=self.bucket, is_upgrade=self.is_upgrade,
                            user_management_type=self.user_mgm_type, user_password=self.user_password)
            postInitPlan.friendly_name = "Post Cluster Init."
            postInitPlan.fail_plan = True
            postInitPlan.report_callback = self.plan_utils.create_callback_procedure("starting", stage="Post Cluster Init", instance_type=self.get_instance_type())
            plan.append(postInitPlan)

        expatProcCondition = True
        if self.is_upgrade:
            expat_check = self.plan_utils.get_procedure_check_file(Expat.__name__)
            expatProcCondition = self.plan_utils.check_file_exists(expat_check)
            if not expatProcCondition and self.plan_utils.check_file_exists(post_init_check):
                logger.info("%s - Adding procedure %s", self.name, Expat.__name__)
                expatProc = Expat(Expat.__name__, self.config, check_file=expat_check, source_version=self.source_version, upgrade_version=self.version,
                 cloud_provider=self.cloud_provider, search_domain=self.search_domain, node_id=self.node_id, default_registry_url=self.default_registry_url,
                 python_image_name=self.python_image_name, managed_containers=bool(self.user_registry_url and self.main_registry_url))
                expatProc.friendly_name = "Post Cluster Upgrade Init"
                expatProc.fail_plan = True
                expatProc.report_callback = self.plan_utils.create_callback_procedure("starting", stage="Post Cluster Upgrade Init", instance_type=self.get_instance_type())
                plan.append(expatProc)
        
        if self.plan_utils.check_file_exists(post_init_check) and expatProcCondition:
            self.add_user_init_script(plan)

        user_init_check = self.plan_utils.get_procedure_check_file(UserInitScript.__name__)
        if self.plan_utils.check_file_exists(user_init_check) and self.plan_utils.check_file_exists(post_init_check) and expatProcCondition:
            self.add_report_all_started(plan)
            self.add_cleanup_pre_init_check_file(plan)

        return plan

class WorkerPostInitPlan(BasePostInitPlan):
    def init(self, name: str, config: ConfigParser, **kwargs: str) -> None:
        super().init(name, config, **kwargs)
        self.instance_type = "worker"
     
    def create(self) -> Sequence[Procedure]:    
        plan: List[Procedure] = []

        self.add_user_init_script(plan)

        user_init_check = self.plan_utils.get_procedure_check_file(UserInitScript.__name__)
        if self.plan_utils.check_file_exists(user_init_check):
            self.add_report_all_started(plan)

        return plan
