import logging
import configparser
import os
import shutil

from typing import Sequence, List

from ec2init import Procedure
from ec2init.plans.init.base_init_plan import BaseInitPlan, MasterBaseInitPlan
from ec2init.plans.preinit.pre_inits import LocalhostPreInit, MasterPreInit, WorkerPreInit
from ec2init.plans.preinit.database_init import LocalhostInitializeDatabase, MasterInitializeDatabase, MasterWithNdbInitializeDatabase
from ec2init.plans.preinit.grow_data_volume import FixBackupDataVolume, GrowDataVolume
from ec2init.plans.init.check_s3_permissions import CheckS3Permissions
from ec2init.utils.backup_information import BackupInformation

logger = logging.getLogger(__name__)

class LocalhostPreInitPlan(MasterBaseInitPlan):
    def init(self, name: str, config: configparser.ConfigParser, **kwargs: str) -> None:
        super().init(name, config, **kwargs)

    def create(self) -> Sequence[Procedure]:
        plan: List[Procedure] = []

        self.add_centos_configure_dns(plan)

        # Make a callback to Hopsworks Cloud indicating the instance has booted
        self.add_callback_instance_booted(plan)

        parentPlan: Sequence[Procedure] = super().create()
        plan += list(parentPlan)

        grow_data_vol_check = self.plan_utils.get_procedure_check_file(GrowDataVolume.__name__)
        init_db_check_file = self.plan_utils.get_procedure_check_file(LocalhostInitializeDatabase.__name__)
        if not self.plan_utils.check_file_exists(init_db_check_file)\
            and self.plan_utils.check_file_exists(grow_data_vol_check):
            logger.info("%s - Adding procedure %s", self.name, LocalhostInitializeDatabase.__name__)
            init_db_task = LocalhostInitializeDatabase(LocalhostInitializeDatabase.__name__, self.config, init_db_check_file,
                self.bucket, self.region, self.ldap_conf_raw, self.oauth_conf_raw, self.password_file, self.user_email,
                self.gateway, self.uid, self.instance_id, self.api_key, self.cloud_provider, self.is_upgrade, self.search_domain, self.node_id, self.user_mgm_type, self.source_version)
            init_db_task.friendly_name = "Initializing database"
            init_db_task.fail_plan = True
            init_db_task.report_callback = self.plan_utils.create_callback_procedure("starting", stage="Initializing database", instance_type = self.instance_type)
            plan.append(init_db_task)

        preinit_check_file = self.plan_utils.get_procedure_check_file(LocalhostPreInit.__name__)
        if not self.plan_utils.check_file_exists(preinit_check_file)\
            and self.plan_utils.check_file_exists(init_db_check_file):
            logger.info("%s - Adding procedure %s", self.name, LocalhostPreInit.__name__)
            preinit_task = LocalhostPreInit(LocalhostPreInit.__name__, self.config, preinit_check_file, self.bucket,
                self.region, self.ldap_conf_raw, self.oauth_conf_raw, self.password_file, self.user_email, self.gateway, self.uid,
                self.instance_id, self.api_key, self.cloud_provider, self.is_upgrade, self.search_domain, self.node_id, self.user_mgm_type)
            preinit_task.friendly_name = "Pre Cluster Init"
            preinit_task.fail_plan = True
            preinit_task.report_callback = self.plan_utils.create_callback_procedure("starting", stage="Preparing instance", instance_type = self.instance_type)
            plan.append(preinit_task)
        return plan

class FixBackupDataVolumePlan(MasterBaseInitPlan):
    def init(self, name: str, config: configparser.ConfigParser, **kwargs: str) -> None:
        super().init(name, config, **kwargs)
        self.instance_type = "master"

    def create(self) -> Sequence[Procedure]:
        plan: List[Procedure] = []
        self.add_fix_backup_data_volume(plan)
        return plan


    def add_fix_backup_data_volume(self, plan: List[Procedure]) -> None:
        check_file = self.plan_utils.get_procedure_check_file(FixBackupDataVolume.__name__)
        if not self.plan_utils.check_file_exists(check_file):
            logger.info(f"{self.name} - Adding procedure {FixBackupDataVolume.__name__}")
            fix_vol = FixBackupDataVolume(FixBackupDataVolume.__name__, self.config, self.plan_utils)
            fix_vol.friendly_name = "Fixing backup data volume mount point"
            fix_vol.fail_plan = True
            plan.append(fix_vol)

class MasterPreInitPlan(MasterBaseInitPlan):
    def init(self, name: str, config: configparser.ConfigParser, **kwargs: str) -> None:
        super().init(name, config, **kwargs)
        self.user_mgm_type = kwargs.get("user_management_type", "")
        if BackupInformation.is_backup(self.backup_info):
            # We are restoring a multidisk backup
            # we need to delete some more check files
            if self.plan_utils.check_file_exists(self.plan_utils.get_procedure_check_file(GrowDataVolume.__name__)):
                files_to_keep = [
                    GrowDataVolume.__name__,
                    CheckS3Permissions.__name__,
                    "backup_info"
                ]
                procedures_check_directory = self.plan_utils.get_procedures_check_directory()
                for file in os.listdir(procedures_check_directory):
                    if file in files_to_keep:
                        continue
                    full_path = os.path.join(procedures_check_directory, file)
                    logger.info(f"Deleting file {full_path}")
                    if os.path.isfile(full_path):
                        os.remove(full_path)
                    else:
                        shutil.rmtree(full_path)
    
    def create(self) -> Sequence[Procedure]:
        plan: List[Procedure] = []

        self.add_centos_configure_dns(plan)

        # Make a callback to Hopsworks Cloud indicating the instance has booted
        self.add_callback_instance_booted(plan)

        parentPlan: Sequence[Procedure] = super().create()
        plan += list(parentPlan)

        grow_data_vol_check = self.plan_utils.get_procedure_check_file(GrowDataVolume.__name__)
        init_db_check_file = self.plan_utils.get_procedure_check_file(MasterInitializeDatabase.__name__)
        if not self.plan_utils.check_file_exists(init_db_check_file)\
            and self.plan_utils.check_file_exists(grow_data_vol_check):
            if self.mgmd_ip == "":
                logger.info("%s - Adding procedure %s", self.name, MasterInitializeDatabase.__name__)
                init_db_task = MasterInitializeDatabase(MasterInitializeDatabase.__name__, self.config, init_db_check_file,
                    self.bucket, self.region, self.ldap_conf_raw, self.oauth_conf_raw, self.password_file, self.user_email,
                    self.gateway, self.uid, self.instance_id, self.api_key, self.cloud_provider, self.is_upgrade, self.search_domain, self.node_id, self.user_mgm_type, self.source_version)
            else:
                logger.info("%s - Adding procedure %s", self.name, MasterWithNdbInitializeDatabase.__name__)
                ## Check file intentionally is the same for both Procedures
                init_db_task = MasterWithNdbInitializeDatabase(MasterWithNdbInitializeDatabase.__name__, self.config, init_db_check_file,
                    self.bucket, self.region, self.ldap_conf_raw, self.oauth_conf_raw, self.password_file, self.user_email,
                    self.gateway, self.uid, self.instance_id, self.api_key, self.cloud_provider,
                    self.mgmd_ip, self.is_upgrade, self.search_domain, self.node_id, self.user_mgm_type, self.source_version)
            init_db_task.friendly_name = "Initializing database"
            init_db_task.fail_plan = True
            init_db_task.report_callback = self.plan_utils.create_callback_procedure("starting", stage="Initializing database", instance_type = self.instance_type)
            plan.append(init_db_task)

        preinit_check_file = self.plan_utils.get_procedure_check_file(MasterPreInit.__name__)
        if not self.plan_utils.check_file_exists(preinit_check_file)\
            and self.plan_utils.check_file_exists(init_db_check_file):
            logger.info("%s - Adding procedure %s", self.name, MasterPreInit.__name__)
            preinit_task = MasterPreInit(MasterPreInit.__name__, self.config, preinit_check_file, self.bucket,
                self.region, self.ldap_conf_raw, self.oauth_conf_raw, self.password_file, self.user_email, self.gateway, self.uid,
                self.instance_id, self.api_key, self.cloud_provider, self.user_mgm_type, self.is_upgrade, self.source_version, self.version, self.search_domain, self.node_id)
            preinit_task.friendly_name = "Master node pre init"
            preinit_task.fail_plan = True
            preinit_task.report_callback = self.plan_utils.create_callback_procedure("starting", stage="Preparing instance", instance_type = self.instance_type)
            plan.append(preinit_task)

        return plan

class WorkerPreInitPlan(BaseInitPlan):
    def init(self, name: str, config: configparser.ConfigParser, **kwargs: str) -> None:
        super().init(name, config, **kwargs)
        self.instance_type = "worker"

    def create(self) -> Sequence[Procedure]:
        plan: List[Procedure] = []

        self.add_centos_configure_dns(plan)

        # Make a callback to Hopsworks Cloud indicating the instance has booted
        self.add_callback_instance_booted(plan)

        parentPlan: Sequence[Procedure] = super().create()
        plan += list(parentPlan)

        preinit_check_file = self.plan_utils.get_procedure_check_file(WorkerPreInitPlan.__name__)
        if not self.plan_utils.check_file_exists(preinit_check_file):
            logger.info("%s - Adding procedure %s", self.name, WorkerPreInitPlan.__name__)
            worker_preinit_task = WorkerPreInit(WorkerPreInit.__name__, self.config, preinit_check_file,
                self.bucket, self.region, self.api_key, self.cloud_provider, self.search_domain, self.node_id)
            worker_preinit_task.friendly_name = "Worker node pre init"
            worker_preinit_task.fail_plan = True
            plan.append(worker_preinit_task)
        return plan

class RonDBPreInitPlan(BaseInitPlan):
    def init(self, name: str, config: configparser.ConfigParser, **kwargs: str) -> None:
        super().init(name, config, **kwargs)

    def create(self) -> Sequence[Procedure]:
        plan: List[Procedure] = []

        self.add_centos_configure_dns(plan)
        self.add_callback_instance_booted(plan, user_email=self.user_email)
        self.add_grow_data_volume(plan)
        parentPlan: Sequence[Procedure] = super().create()
        plan += list(parentPlan)
        return plan

class MgmRonDBPreInitPlan(RonDBPreInitPlan):
    def init(self, name: str, config: configparser.ConfigParser, **kwargs: str) -> None:
        super().init(name, config, **kwargs)
        self.instance_type = "ndb_mgm"

class NdbdRonDBPreInitPlan(RonDBPreInitPlan):
    def init(self, name: str, config: configparser.ConfigParser, **kwargs: str) -> None:
        super().init(name, config, **kwargs)
        self.instance_type = "ndb_ndbd"

class MysqldRonDBPreInitPlan(RonDBPreInitPlan):
    def init(self, name: str, config: configparser.ConfigParser, **kwargs: str) -> None:
        super().init(name, config, **kwargs)
        self.instance_type = "mysqld"

class ApiRonDBPreInitPlan(RonDBPreInitPlan):
    def init(self, name: str, config: configparser.ConfigParser, **kwargs: str) -> None:
        super().init(name, config, **kwargs)
        self.instance_type = "ndb_api"