from ec2init.utils.database import databaseUtils
import logging
import configparser
import pathlib
import subprocess
import os
import xml.etree.ElementTree as ET
import json
import time
import glob
import shutil
import base64


from pathlib import Path

from typing import List, Tuple, Optional

from ec2init import Procedure
from ec2init.utils.system import OSType, SystemUtils
from ec2init.utils.cloud import get_cloud_provider, Provider
from ec2init.plans.plan_utils import PlanUtils, HopsPlanUtils
from ec2init.utils.http import HttpUtils
from ec2init.plans.init.generate_hops_certs import GenerateHopsCerts
from ec2init.plans.multinode.master_seds import MasterSeds
from ec2init.plans.multinode.register_generate_certs import RegisterGenerateCerts
from ec2init.plans.postinit.store_backup_info import StoreBackupInfo
from ec2init.plans.init.callback_hopsworks import CallbackHopsworks
from ec2init.plans.init.load_docker import LoadDocker
from ec2init.plans.init.setup_eks import SetupEKS
from ec2init.plans.init.setup_ecr_docker_images import SetupECRDockerImages
from ec2init.utils.hopsfs import HopsFSUtils
from ec2init.utils.backup_information import BackupInformation

logger = logging.getLogger(__name__)

MYSQL_CLIENT = "/srv/hops/mysql-cluster/ndb/scripts/mysql-client.sh"

class BasePreInit(Procedure):
    def __init__(self, name: str, config: configparser.ConfigParser, check_file: pathlib.Path,
        bucket: str, region: str, cloud_provider: str, search_domain: str, node_id: str):
        self.name = name
        self.config = config
        self.check_file = check_file
        self.bucket = bucket
        self.region = region
        self.cloud_provider = cloud_provider
        self.search_domain = search_domain
        self.node_id = node_id

    def pre(self) -> None:
        if self.report_callback is not None:
            try:
                self.report_callback.execute()
            except Exception as ex:
                logger.warning("%s - Could not execute pre-execute: %s", self.name, ex)

    def execute(self) -> List[Procedure]:
        return []

    def rollback(self) -> None:
        pass

    def get_local_name(self) -> str:
        try:
            return get_cloud_provider(self.cloud_provider).get_fqdn(self.config, self.search_domain, self.node_id)
        except Exception as ex:
            self.friendly_name = str(ex)
            raise ex

    def get_local_ip(self) -> str:
        try:
            return get_cloud_provider(self.cloud_provider).get_private_ip(self.config)
        except Exception as ex:
            self.friendly_name = "Cluster initialization failed, could not get local IP address"
            raise ex

    def sysexec(self, cmd: List[str], timeout: int = 120) -> subprocess.CompletedProcess:
        return SystemUtils.execute_command(cmd, timeout=timeout)

    def fix_hopsfs_s3_conf_attrs(self) -> None:

        if self.bucket:
            HopsFSUtils.fix_hdfs_site("dfs.replication", "1")
            HopsFSUtils.fix_hdfs_site("dfs.enable.cloud.persistence", "true"); 
            HopsFSUtils.fix_hdfs_site("dfs.cloud.aws.s3.region", self.region); 
            HopsFSUtils.fix_hdfs_site("dfs.cloud.aws.s3.bucket", self.bucket); 
            HopsFSUtils.fix_hdfs_site("dfs.disable.non.cloud.storage.policies", "true"); 
        else:
            HopsFSUtils.fix_hdfs_site("dfs.enable.cloud.persistence", "false"); 
            HopsFSUtils.fix_hdfs_site("dfs.cloud.aws.s3.region", ""); 
            HopsFSUtils.fix_hdfs_site("dfs.cloud.aws.s3.bucket", ""); 
    
    def enable_installed_gpus(self) -> None:
        if self.gpu_installed() and self.install_nvidia_driver() and self.install_nvidia_packages():
            #change config files only if all packages are installed successfully
            self.fix_yarn_site_for_gpus()
            self.fix_container_executor_for_gpus()
            self.fix_docker_runtime()
            self.configure_nvml_monitor()

    def gpu_installed(self) -> bool:
        try:
            SystemUtils.execute_command([f"lspci | grep NVIDIA"], timeout=120, shell=True)
            logger.info("NVIDIA GPU found.")
            return True
        except subprocess.CalledProcessError:
            return False
    
    def install_nvidia_driver(self) -> bool:
        try:
            driver_download_url = self.config.get('initplan', 'gpu_driver_url', fallback=None)
            if not driver_download_url:
                raise Exception('Missing Nvidia driver download URL. Check your configuration!')
            driver = os.path.basename(driver_download_url)
            process = SystemUtils.execute_command([f"wget {driver_download_url} && chmod +x {driver} && ./{driver}  -a --install-libglvnd --force-libglx-indirect -q --compat32-libdir -s --ui=none && rm -f {driver}"], timeout=300, shell=True)
            logger.info("Nvidia Driver Installed. Ret Code: "+ str(process.returncode))
            return True
        except Exception:
            message = "Unable to install NVIDIA drivers"
            self.friendly_name = message 
            raise Exception(message)

    def fix_yarn_site_for_gpus(self) -> None:
        yarn_site_xml = "/srv/hops/hadoop/etc/hadoop/yarn-site.xml"
        prop_name = "yarn.nodemanager.resource-plugins"
        prop_value = "yarn.io/gpu" 
        found = False

        tree = ET.parse(yarn_site_xml)
        root = tree.getroot()

        for p in root.iter('property'):
            if p[0].text == prop_name:
                logger.warn(f"{prop_name} is already defined in the yarn-site.xml with value: {p[1].text}")
                p[1].text = prop_value; #overwrite the value
                found = True
                tree.write(open(yarn_site_xml, 'w'), encoding='unicode')
                break
        
        if not found:
            new_child = ET.Element("property")
            new_child.append(ET.Element("name"))
            new_child.append(ET.Element("value"))
            new_child[0].text = prop_name 
            new_child[1].text = prop_value 
            root.append(new_child)
            tree.write(open(yarn_site_xml, 'w'), encoding='unicode')

    def fix_container_executor_for_gpus(self) -> None:
        container_exe_cfg = "/srv/hops/hadoop/etc/hadoop/container-executor.cfg"

        f = open(container_exe_cfg, "r")
        contents = f.readlines()
        f.close()

        i = 0
        search_line = "docker.allowed.volume-drivers="
        while i < len(contents):
            if contents[i].strip().startswith(search_line):
                break
            i += 1

        driver_download_url = self.config.get('initplan', 'gpu_driver_url', fallback=None)
        if not driver_download_url:
            raise Exception('Missing Nvidia driver download URL. Check your configuration!')
        version = driver_download_url[driver_download_url.index("x86_64-") + len("x86_64-"):-4]

        new_prop = ("docker.allowed.volume-drivers=nvidia-docker\n"
                    f"docker.allowed.ro-mounts=nvidia_driver_{version}\n"
                    "docker.allowed.runtimes=nvidia\n"
                    "[gpu]\n"
                    "module.enabled=true\n"
                    "[cgroups]\n"
                    "# This should be same as yarn.nodemanager.linux-container-executor.cgroups.mount-path inside yarn-site.xml\n"
                    "root=/sys/fs/cgroup\n"
                    "# This should be same as yarn.nodemanager.linux-container-executor.cgroups.hierarchy inside yarn-site.xml\n"
                    "yarn-hierarchy=hops-yarn\n")

        contents.pop(i)
        contents.append(new_prop)
        
        f = open(container_exe_cfg, "w")
        newcontents = "".join(contents)
        f.write(newcontents)
        f.close()
    
    def fix_docker_runtime(self) -> None:
        daemon_json = "/etc/docker/daemon.json"
        with open(daemon_json) as f:
            data = json.load(f)
            data["default-runtime"] = "nvidia"
            if not "runtimes" in data:
                data["runtimes"] = {}

            data["runtimes"]["nvidia"] = {}
            data["runtimes"]["nvidia"]["path"] = "nvidia-container-runtime"
            data["runtimes"]["nvidia"]["runtimeArgs"] = [] 

        with open(daemon_json, 'w') as outfile:
            json.dump(data, outfile, indent=4)
        
        #Docker is running at this stage. Restart it to read the new changes
        SystemUtils.execute_command(["systemctl", "restart", "docker"], timeout=120)

    def install_nvidia_packages(self) -> bool:
        for _ in range(0, 3):                
            try:
                #rename the /etc/docker/daemon.json file
                #otherwise installing the following packages shows (y/n) prompt
                #to overwrite this file. The file will be restored later
                daemon_json = "/etc/docker/daemon.json"
                daemon_json_bak = "/etc/docker/daemon.json.bak"
                os.rename(daemon_json, daemon_json_bak)

                nvidia_pkgs_url = self.config.get('initplan', 'nvidia_pkgs_url', fallback=None)
                if not nvidia_pkgs_url:
                    raise Exception('Missing Nvidia packages download URL. Check your configuration!')
                packages = [ "libnvidia-container1_1.0.7-1_amd64.deb",
                              "nvidia-container-runtime_3.1.4-1_amd64.deb",
                              "libnvidia-container-tools_1.0.7-1_amd64.deb",
                              "nvidia-container-toolkit_1.0.5-1_amd64.deb",
                              "nvidia-docker2_2.2.2-1_all.deb" ]
                for package in packages:
                    url = f"{nvidia_pkgs_url}/{package}"
                    process = SystemUtils.execute_command([f"wget {url}"], timeout=120, shell=True)
                    logger.info("Downloaded "+package+" Ret Code: "+ str(process.returncode))

                install_cmd  = "dpkg -i "+ " ".join(packages) 
                SystemUtils.execute_command([f"{install_cmd}"], timeout=120, shell=True)

                remove_cmd  = "rm -f "+ " ".join(packages) 
                SystemUtils.execute_command([f"{remove_cmd}"], timeout=120, shell=True)

                logger.info("Installed Docker packages for NVIDIA")

                return True
            except subprocess.CalledProcessError:
                logger.error("Unable to Install Docker packages for NVIDIA")
            finally:
                os.rename(daemon_json_bak, daemon_json)
            
            logger.info("Retrying Installing Docker packages for NVIDIA")
            time.sleep(10)
            
        message = "Unable to install Docker packages for NVIDIA"
        self.friendly_name = message 
        raise Exception(message)

    def configure_nvml_monitor(self) -> None:
        logger.info("Enabling GPU monitoring service")
        systemd_unit_name: str = "nvml_monitor"
        commands: List[List[str]] = [
            ["systemctl", "enable", systemd_unit_name],
            ["systemctl", "daemon-reload"],
            ["systemctl", "start", systemd_unit_name]
        ]
        [SystemUtils.execute_command(cmd) for cmd in commands]
        logger.info("Enabled GPU monitoring service")
        logger.debug("Updating kagent services list")
        kagent_conf_dir = self.config.get(
            'general', 'kagent_conf_dir', fallback="/srv/hops/kagent/etc")
        services_file = pathlib.Path(kagent_conf_dir, "services")
        services = configparser.ConfigParser()
        services.read(str(services_file))
        service_group = "Monitoring"
        service_name = "nvml_monitor"
        service_section = f"{service_group}-{service_name}"
        if service_section not in services:
            services[service_section] = {
                'group': service_group,
                'service': service_name,
                'stdout-file': '',
                'config-file': '',
                'fail-attempts': '1'
            }
            with open(str(services_file), 'w') as fd:
                services.write(fd)
            logger.info(f"Added {service_name} to host services list")
        else:
            logger.debug(f"Service {service_section} already exist in host services list")

    def sed_stmt(self, regex: str, file: str) -> str:
        return f"sed -i '{regex}' {file}"

    def sed_cmd(self, regex: str, file: str) -> List[str]:
        return ["sh", "-c", self.sed_stmt(regex, file)]

    def touch_check_file(self) -> None:
        if self.check_file is not None:
            self.check_file.touch()

    def update_kagent_password(self, password) -> None:
        self._config_update_kagent_password(password)

    def _config_update_kagent_password(self, password) -> None:
        kagent_conf_dir = self.config.get(
            'general', 'kagent_conf_dir', fallback="/srv/hops/kagent/etc")
        kagent_conf = Path(kagent_conf_dir, "config.ini")
        self.sysexec(self.sed_cmd(f"s/password[[:space:]]=.*/password = {password}/g", str(kagent_conf)))


class BrainsPreInit(BasePreInit):
    def __init__(self, name: str, config: configparser.ConfigParser, check_file: pathlib.Path,
        bucket: str, region: str, ldap: str, oauth: str, password_file: pathlib.Path, user_email: str,
        gateway: str, user_id: str, instance_id: str, api_key: str, cloud_provider: str, is_upgrade: bool, search_domain: str, node_id: str, user_mgm_type: str):
        super().__init__(name, config, check_file, bucket, region, cloud_provider, search_domain, node_id)
        self.ldap = ldap
        self.oauth = oauth
        self.password_file = password_file
        self.user_email = user_email
        self.gateway = gateway
        self.user_id = user_id
        self.instance_id = instance_id
        self.api_key = api_key
        self.is_upgrade = is_upgrade
        self.plan_utils = PlanUtils(PlanUtils.create_checks_directory(config))
        backup_info = self.plan_utils.get_procedure_check_file("backup_info")
        logger.info("backup_info path " + str(backup_info))
        self.backup_info = BackupInformation(config)
        if self.plan_utils.check_file_exists(backup_info):
            self.backup_info.load(pathlib.Path(backup_info))
        logger.info("is_backup: " + str(BackupInformation.is_backup(self.backup_info)))
        self.user_mgm_type = user_mgm_type
        if BackupInformation.is_backup(self.backup_info):
            kagent_conf_dir = self.config.get('general', 'kagent_conf_dir', fallback="/srv/hops/kagent/etc")
            services_file = pathlib.Path(kagent_conf_dir, "services")
            services = configparser.ConfigParser()
            services.read(str(services_file))
            services_to_skip = [
                'service_does_not_exist',
                'glassfish-domain1',
                'docker',
                # Skip database. The configuration is updated and the database is restarted in database_init.py
                'ndb_mgmd',
                'ndbmtd',
                'mysqld'
            ]
            for section in services.sections():
                service = services.get(section, 'service', fallback='service_does_not_exist')
                #do not stop
                if service not in services_to_skip:
                    logger.debug("%s - Stopping service: %s", self.name, service)
                    self._stop_service(service)
                    logger.info("%s - Stopped service: %s", self.name, service)
            lets_encrypt_dir = self.config.get('initplan', 'lets_encrypt_dir', fallback="/etc/letsencrypt")
            script = pathlib.Path(lets_encrypt_dir, 'renewal-hooks', 'deploy', 'deploy2glassfish_hook.sh')
            if script.is_file() :
                script.unlink()
            SystemUtils.execute_command(["systemctl", "stop", "consul"], timeout=120)
            SystemUtils.execute_command(["systemctl", "disable", "ec2update"], timeout=120)

            
    def substitute_in_db(self, local_name: str, local_ip: str) -> None:
        subs: List[Tuple[str, str]] = [
            ("hopsworks_endpoint", local_name),
            ("yarn_ui_ip", local_ip),
            ("elastic_ip", local_ip),
            ("kibana_ip", local_ip),
            ("localhost", "false")
        ]

        if not self.is_upgrade:
            ldap_enabled = "true" if self.ldap else "false"
            oauth_enabled = "true" if self.oauth else "false"
            subs.append(("ldap_auth", ldap_enabled))
            subs.append(("oauth_enabled", oauth_enabled))


        for s in subs:
            k, v = s
            databaseUtils.update_variable(k, v)

    def fix_hopsfs_s3_db_attrs(self) -> None:
        if self.bucket:
            databaseUtils.run_command(["hops", "-e", "truncate hdfs_cloud_buckets"])
            databaseUtils.run_command(["hops", "-e", f"insert into hdfs_cloud_buckets values (1, '{self.bucket}')"])
            databaseUtils.run_command(["hops", "-e", f"UPDATE hdfs_inodes SET storage_policy=3 where is_dir = 1 and storage_policy != 0"])
            databaseUtils.run_command(["hops", "-e", f"UPDATE hdfs_inodes SET storage_policy=3 WHERE id=1"])

            databaseUtils.update_variable("hdfs_base_storage_policy", "CLOUD")
            databaseUtils.update_variable("hdfs_log_storage_policy", "CLOUD")
        else:
            databaseUtils.run_command(["hops", "-e", "truncate hdfs_cloud_buckets"])
            databaseUtils.run_command(["hops", "-e", f"UPDATE hdfs_inodes SET storage_policy=7 WHERE id=1"])

            databaseUtils.update_variable("hdfs_base_storage_policy", "DB")
            databaseUtils.update_variable("hdfs_log_storage_policy", "HOT")

    def update_hopsfs_time_stamps(self) -> None:
           ts = int(round(time.time() * 1000))
           logger.info("Setting inodes and block time stamp to " + str(ts))
           databaseUtils.run_command(["hops", "-e", f"UPDATE hdfs_inodes SET access_time = {ts}, modification_time = {ts}"])
           databaseUtils.run_command(["hops", "-e", f"UPDATE hdfs_block_infos SET time_stamp = {ts}"])

    def generate_new_jwt_signing_key(self) -> None:
        if BackupInformation.is_backup(self.backup_info):
            return
        logger.info("Replacing JWT signing key")
        rand = os.urandom(64)
        encoded_rand = base64.b64encode(rand).decode('utf-8')
        sql_update = f"REPLACE INTO jwt_signing_key (secret, name) values ('{encoded_rand}', 'apiKey')"
        databaseUtils.run_command(["hopsworks", "-e", sql_update])

    def create_hopsworks_user(self) -> None:
        if BackupInformation.is_backup(self.backup_info) or self.is_upgrade:
            #if we are recovering from a backup the user should already exist in the database
            #also if we are upgrading, since this will be handled by the upgrade_cluster proc
            logger.info("Skip creating new Hopsworks user during backup restore and upgrade")

            # We may still need to do some call with the admin user. For this purpose we reactivate it
            # and update its password, it will be deactivated at the end of the process.
            databaseUtils.run_command(["hopsworks", "-e","UPDATE users SET status=2 where username='meb10000'"])
            databaseUtils.update_variable("disable_password_login", "false")

            process=self.sysexec(["openssl", "rand", "-base64", "32"])
            password=process.stdout.strip().decode("utf-8")
            with open(self.password_file, "w") as fd:
                fd.write(password)

            salt=self.get_existing_user_salt("meb10000")
            sha=self.password_digest(password, salt)
            databaseUtils.run_command(["hopsworks", "-e",f"UPDATE users SET password='{sha}' where username='meb10000'"])
            return
        if self.config.getboolean('general', 'unmanaged', fallback=False):
            logger.warning("Skip creating new Hopsworks user because it's an unmanaged AMI")
            return
        process = self.sysexec(["openssl", "rand", "-base64", "32"])
        password = process.stdout.strip().decode("utf-8")
        with open(self.password_file, "w") as fd:
            fd.write(password)

        salt = self.get_existing_user_salt("meb10000")
        sha = self.password_digest(password, salt)
        
        if self.user_mgm_type.upper() != 'MANAGED':
            process = SystemUtils.execute_command(["/srv/hops/mysql-cluster/ndb/scripts/mysql-client.sh hopsworks -e 'select max(uid) from users' -sN | tail -1"],
                 timeout=120, shell=True)
            uuid = process.stdout.strip().decode("utf-8")
            uuid = str(int(uuid) + 1)
            databaseUtils.run_command(["hopsworks",  "-e",
                 "INSERT INTO users (uid, username, password, email, fname, lname, activated, title, orcid, false_login, isonline, secret, validation_key, mode, password_changed, notes, status, max_num_projects, two_factor, tours_state , salt)" +
                f"VALUES ('{uuid}', 'meb{uuid}', '{sha}', '{self.user_email}', 'Admin', 'Admin', '2015-05-15 10:22:36', 'Mx', '1234-5678-1234-5678', -1, 0, 'V3WBPS4G2WMQ53VA', NULL, 0, '2015-04-28 15:18:42', NULL, 2, 30, 0, 3, '{salt}')"])

            databaseUtils.run_command(["hopsworks", "-e",
                 f"INSERT INTO account_audit (initiator, target, action, action_timestamp, message, outcome, ip, useragent) values ({uuid}, {uuid}, 'REGISTRATION', CURRENT_TIMESTAMP, 'Admin account was enabled', 'SUCCESS', '127.0.0.1', 'Firefox')"])

            databaseUtils.run_command(["hopsworks", "-e",
                 f"REPLACE INTO user_group (uid, gid) VALUES ({uuid}, 1005)"])

        databaseUtils.run_command(["hopsworks", "-e",
             f"UPDATE users SET password='{sha}' where username='meb10000'"])

    def update_kagent_password(self, password) -> None:
        self._db_update_kagent_password(password)
        self._config_update_kagent_password(password)
    
    def update_onlinefs_password(self, password) -> None:
        self._db_update_onlinefs_password(password)

    def _db_update_kagent_password(self, password) -> None:
        if BackupInformation.is_backup(self.backup_info):
            return
        self._db_update_user_password("agent007", password)

    def _db_update_onlinefs_password(self, password) -> None:
        if BackupInformation.is_backup(self.backup_info):
            return
        self._db_update_user_password("onlinefs", password)
    
    def _db_update_user_password(self, username, password) -> None:
        salt = self.get_existing_user_salt(username)
        sha = self.password_digest(password, salt)
        sql_update = f"UPDATE users SET password='{sha}' where username='{username}'"
        databaseUtils.run_command(["hopsworks", "-e", sql_update])

    def get_existing_user_salt(self, username) -> str:
        process = SystemUtils.execute_command([f"/srv/hops/mysql-cluster/ndb/scripts/mysql-client.sh hopsworks -e \"select salt from users where username='{username}'\" -sN | tail -1"], timeout=120, shell=True)
        return process.stdout.strip().decode("utf-8")

    def password_digest(self, password, salt) -> str:
        salted_password = password + salt
        process = SystemUtils.execute_command([f"echo -n {salted_password} | sha256sum | cut -d \" \" -f 1"], timeout=120, shell=True)
        return process.stdout.strip().decode("utf-8")

    def update_cloud_callback(self) -> None:
        # add cloud_events_endpoint to the variables table"
        databaseUtils.run_command(["hopsworks", "-e",
             f"REPLACE INTO variables (id, value) values ('cloud_events_endpoint', '{self.gateway}/eventsv2/ops/users/{self.user_id}/instances/{self.instance_id}')"])

        databaseUtils.run_command(["hopsworks", "-e",
             f"REPLACE INTO variables (id, value) values ('cloud_events_endpoint_api_key', '{self.api_key}')"])

    def restart_spark_flink_history_servers(self) -> None:
        #flink and spark history server are already started by systemd and need to restart to use new conf
        self.sysexec(["systemctl", "stop", "flinkhistoryserver"])
        self.sysexec(["systemctl", "stop", "sparkhistoryserver"])

    def _stop_service(self, service: str) -> None:
        try:
            SystemUtils.execute_command(["systemctl", "stop", service], timeout=120)
        except Exception as ex:
            logger.error("%s - Could not stop service: %s", service, ex)
    
    def deny_outside_access_to_yarn_ui(self) -> None:
        logger.info("Deny outside access to YARN")
        try:
            SystemUtils.execute_command(["iptables", "-A", "INPUT", "-i", "lo", "-p", "tcp", "--dport", "8088", "-j", "ACCEPT"])
            SystemUtils.execute_command(["iptables", "-A", "INPUT", "-p", "tcp", "--dport", "8088", "-j", "DROP"])
            os_type = SystemUtils.get_os_type()
            logger.info(f"{self.name} - OS type is: {os_type}")
            if os_type == OSType.UBUNTU:
                SystemUtils.execute_command(["iptables-save | tee /etc/iptables/rules.v4"], shell=True)
                SystemUtils.execute_command(["ip6tables-save | sudo tee /etc/iptables/rules.v6"], shell=True)
            elif os_type == OSType.CENTOS:
                cmd = "yum -y install iptables-services"
                SystemUtils.execute_command(cmd.split(" "), timeout=180)
                cmd = "systemctl enable iptables"
                SystemUtils.execute_command(cmd.split(" "))
                cmd = "/usr/libexec/iptables/iptables.init save"
                SystemUtils.execute_command(cmd.split(" "))
            logger.info(f"{self.name} - Finished setting iptables rules")
        except Exception as ex: 
            logger.error("Could not add iptables rules to deny access to YARN web ui: %s", ex)
            raise ex
    
class LocalhostPreInit(BrainsPreInit):
    def __init__(self, name: str, config: configparser.ConfigParser, check_file: pathlib.Path,
        bucket: str, region: str, ldap: str, oauth: str, password_file: pathlib.Path, user_email: str,
        gateway: str, user_id: str, instance_id: str, api_key: str, cloud_provider: str, is_upgrade: bool, search_domain: str, node_id: str, user_mgm_type: str):
        super().__init__(name, config, check_file, bucket, region, ldap, oauth, password_file, user_email,
            gateway, user_id, instance_id, api_key, cloud_provider, is_upgrade, search_domain, node_id, user_mgm_type)

    def execute(self) -> List[Procedure]:
        local_name = self.get_local_name()
        local_ip = self.get_local_ip()

        hostHolder = "localhost"
        ipHolder = "127.0.0.1"
        if BackupInformation.is_backup(self.backup_info) and self.backup_info is not None:
            regenerate_cert_check = self.plan_utils.get_procedure_check_file(GenerateHopsCerts.__name__)
            self.plan_utils.remove_check_file(regenerate_cert_check)
            
            if self.backup_info.ip:
                ipHolder = self.backup_info.ip
            if self.backup_info.hostname:
                hostHolder = self.backup_info.hostname
            
        # Change mysqld bind address
        self.sysexec(self.sed_cmd("s/{ipHolder}$/0.0.0.0/g", "/srv/hops/mysql-cluster/my.cnf"))
        
        self.substitute_in_db(local_name, local_ip)

        kagent_conf_dir = self.config.get('general', 'kagent_conf_dir', fallback="/srv/hops/kagent/etc")
        kagent_conf_file = Path(kagent_conf_dir, 'config.ini')
        kagent_conf = configparser.ConfigParser()
        kagent_conf.read(str(kagent_conf_file))
        state_store_location = kagent_conf.get('agent', 'state-store', fallback='/srv/hops/kagent/etc/state_store')
        state_store = os.path.join(state_store_location, "crypto_material_state.json")

        subs: List[Tuple[str, str]] = [
            (f"s/{hostHolder}/{local_name}/g", "/srv/hops/hadoop/etc/hadoop/hdfs-site.xml"),
            (f"s/{ipHolder}/{local_ip}/g", "/srv/hops/hadoop/etc/hadoop/mapred-site.xml"),

            (f"s/{ipHolder}/{local_ip}/g", "/srv/hops/spark/conf/spark-env.sh"),
            (f"s/{ipHolder}/{local_ip}/g", "/srv/hops/spark/conf/spark-defaults.conf"),
            (f"s/{ipHolder}/{local_ip}/g", "/srv/hops/spark/conf/hive-site.xml"),
            (f"s/{ipHolder}:9083/{local_ip}:9083/g", "/srv/hops/spark/conf/*"),

            (f"s/{ipHolder}:8020/{local_ip}:8020/g", "/srv/hops/apache-hive/conf/*"),
            (f"s/{ipHolder}:9083/{local_ip}:9083/g", "/srv/hops/apache-hive/conf/*"),

            (f"s/{ipHolder}:9091/{local_ip}:9091/g", "/srv/hops/kafka/config/server.properties"),
            (f"s/{ipHolder}:9092/{local_ip}:9092/g", "/srv/hops/kafka/config/server.properties"),
            (f"s/User:{hostHolder}/User:{local_name}/g", "/srv/hops/kafka/config/server.properties"),
            (f"s/{ipHolder}$/{local_ip}/g", "/srv/hops/kafka/config/server.properties"),

            (f"s/{hostHolder}:9200/{local_name}:9200/g", "/srv/hops/kibana/config/kibana.yml"),
            (f"s/{hostHolder}/{local_name}/g", "/srv/hops/elasticsearch/config/elasticsearch.yml"),
            (f"s/{hostHolder}/{local_name}/g", "/srv/hops/logstash/config/*"),

            (f"s/{ipHolder}:9200/{local_ip}:9200/g", "/srv/hops/epipe/conf/config.ini"),
            (f"s/{ipHolder}:9200/{local_ip}:9200/g", "/srv/hops/epipe/conf/config-reindex.ini"),


            (f"s/livy\.server\.host[[:blank:]]=[[:blank:]].*/livy\.server\.host = {local_ip}/g", "/srv/hops/apache-livy/conf/livy.conf"),

            (f"s/{hostHolder}/{local_name}/g", "/lib/systemd/system/consul.service"),
            (f"s/{ipHolder}/{local_ip}/g", "/lib/systemd/system/consul.service"),
            (f"s/$(hostname -f | tr -d '[:space:]')/{local_name}/g", "/lib/systemd/system/consul.service"),
            (f"s/{hostHolder}/{local_name}/g", "/srv/hops/consul/.bashrc"),

            (f"s/{hostHolder}/{local_name}/g", str(state_store))

        ]

        for s in subs:
            k, v = s
            self.sysexec(self.sed_cmd(k, v))

        self.sysexec(["systemctl", "daemon-reload"])

        agent_password = self.api_key
        if self.config.getboolean('general', 'unmanaged', fallback=False):
            agent_password = base64.b64encode(os.urandom(32)).decode('utf-8')
        self.fix_hopsfs_s3_conf_attrs()
        self.fix_hopsfs_s3_db_attrs()

        if not BackupInformation.is_backup(self.backup_info) and not self.is_upgrade:
            self.update_hopsfs_time_stamps()

        self.create_hopsworks_user()
        self.update_kagent_password(self.api_key)
        self.generate_new_jwt_signing_key()
        self.update_cloud_callback()
        self.restart_spark_flink_history_servers()
        self.touch_check_file()
        return []

class MasterPreInit(BrainsPreInit):

    def __init__(self, name: str, config: configparser.ConfigParser, check_file: pathlib.Path,
        bucket: str, region: str, ldap: str, oauth: str, password_file: pathlib.Path, user_email: str,
        gateway: str, user_id: str, instance_id: str, api_key: str, cloud_provider: str, user_mgm_type: str,
        is_upgrade: bool, source_version: str, upgrade_version: str, search_domain: str, node_id: str):
        super().__init__(name, config, check_file, bucket, region, ldap, oauth, password_file, user_email,
            gateway, user_id, instance_id, api_key, cloud_provider, is_upgrade, search_domain, node_id, user_mgm_type)
        self.source_version = source_version
        self.upgrade_version = upgrade_version

    def execute(self) -> List[Procedure]:
        local_name = self.get_local_name()
        local_ip = self.get_local_ip()

        if BackupInformation.is_backup(self.backup_info):
            sed_check = self.plan_utils.get_procedure_check_file(MasterSeds.__name__)
            self.plan_utils.remove_check_file(sed_check)
            register_generate_cert_check = self.plan_utils.get_procedure_check_file(RegisterGenerateCerts.__name__)
            self.plan_utils.remove_check_file(register_generate_cert_check)
            
        self.run_flyway_on_upgrade()
        self.substitute_in_db(local_name, local_ip)

        # In case of backup restore we want to remove old nodes
        certs_dir = self.config.get(
            'general', 'certs_dir', fallback="/srv/hops/certs-dir")
            # If possible do not include consul TLD as we might need to change them in the future
        certs_to_revoke = ["L=consul", "L=elastic", "L=elkadmin", "L=flink", "L=glassfish", "L=hdfs",
                           "L=hive", "L=hopsmon", "L=kafka", "L=kagent", "L=livy", "L=mapred", "L=mysql", "L=rmyarn",
                           "L=yarn", "L=onlinefs", "CN=hopsworks.glassfish.service"]
        self.plan_utils.revoke_old_certificates(self.config, certs_dir, certs_to_revoke)
        databaseUtils.run_command(["hopsworks", "-e", f"DELETE FROM hopsworks.hosts"])
        databaseUtils.run_command(["hopsworks", "-e",
             f"INSERT INTO hopsworks.hosts(hostname, host_ip) VALUES ('{local_name}', '{local_ip}') ON DUPLICATE KEY UPDATE cores=NULL;"])

        if not self.is_upgrade:
            # Disable registration (UI)
            if self.user_mgm_type.upper() == 'LDAP':
                databaseUtils.update_variable("disable_registration", "true")
            elif self.user_mgm_type.upper() == 'MANAGED':
                databaseUtils.update_variable("disable_registration", "true")

        self.fix_hopsfs_s3_conf_attrs()
        self.fix_hopsfs_s3_db_attrs()

        if not BackupInformation.is_backup(self.backup_info) and not self.is_upgrade:
            self.update_hopsfs_time_stamps()

        self.fix_yarn_site_for_gpus()
        self.create_hopsworks_user()
        self.update_kagent_password(self.api_key)
        self.update_onlinefs_password(self.api_key)
        self.generate_new_jwt_signing_key()
        self.update_cloud_callback()
        self.restart_spark_flink_history_servers()
        self.deny_outside_access_to_yarn_ui()

        self.touch_check_file()
        return []
    
    def run_flyway_on_upgrade(self) -> None:
        if self.is_upgrade:
            try:
                logger.info("Upgrading HopsFS Schema")
                SystemUtils.execute_command(['/srv/hops/ndb-hops/flyway/flyway', 'migrate'], timeout=300)
                logger.info("Upgrading Hopsworks Schema")
                hopsworks_flyway_dir = '/srv/hops/domains/domain1/flyway'
                SystemUtils.execute_command([f'{hopsworks_flyway_dir}/flyway', 'migrate'], timeout=300)

                # Run all dml files during upgrade
                hopsworks_versions = self.config.get('upgradeplan', 'hopsworks_versions').split(',')
                for version in hopsworks_versions:
                    SystemUtils.execute_command([f'{MYSQL_CLIENT} hopsworks < {hopsworks_flyway_dir}/dml/V{version}__hopsworks.sql'], shell=True, timeout=300)

                # Run upgrade on hive DB
                logger.info("Upgrading Hive Schema")
                hive_schematool = self.config.get('upgradeplan', 'hive_schematool', fallback='/srv/hops/apache-hive/bin/schematool')
                SystemUtils.execute_command([hive_schematool, '-dbType', 'mysql', '-upgradeSchema'], timeout=300)

            except subprocess.CalledProcessError as ex:
                stdout = "None" if ex.stdout is None else ex.stdout.decode("utf-8")
                stderr = "None" if ex.stderr is None else ex.stderr.decode("utf-8")
                logger.error("Failed to run flyway during upgrade: %s\n stdout: %s \nstderr: %s\n", ex, stdout, stderr)
                raise ex
            except Exception as ex:
                logger.error("Failed to run flyway during upgrade: %s", ex)
                raise ex

class WorkerPreInit(BasePreInit):
    def __init__(self, name: str, config: configparser.ConfigParser, check_file: pathlib.Path,
        bucket: str, region: str, api_key:str, cloud_provider: str, search_domain: str, node_id: str):
        super().__init__(name, config, check_file, bucket, region, cloud_provider, search_domain, node_id)
        self.api_key = api_key

    def execute(self) -> List[Procedure]:
        self.fix_hopsfs_s3_conf_attrs()
        self.enable_installed_gpus()
        self.delete_datanode_directories()
        self.update_kagent_password(self.api_key)
        self.touch_check_file()

        return []

    def delete_datanode_directories(self) -> None:
        data_dir = Path(self.config.get('initplan', 'hops_data_dir', fallback='/srv/hops/hopsdata'))
        dn_base_dir = data_dir / "hdfs" / "dn"

        cloud_dn_dir = dn_base_dir / "cloud" / "*"
        dirs_to_delete: List[str] = []
        dirs_to_delete += glob.glob(str(cloud_dn_dir))

        disk_dn_dir =  dn_base_dir / "disk" / "*"
        dirs_to_delete += glob.glob(str(disk_dn_dir))

        for d in dirs_to_delete:
            if os.path.isdir(d):
                shutil.rmtree(d)
            else:
                os.remove(d)
