import abc
import logging
import shutil
import os
import boto3 
from mypy_boto3_ec2 import Client, ServiceResource
import time
from typing import List, Callable, Any
from ec2init import Procedure
from ec2init.utils.system import SystemUtils
from pathlib import Path
from configparser import ConfigParser
from ec2init.utils.cloud import get_cloud_provider
import re

from ec2init.plans.plan_utils import PlanUtils, HopsPlanUtils
from ec2init.plans.upgrade.parallel_cp import ParallelCp, ParallelCpTask
from ec2init.utils.shell import ShellUtils
import docker 
from azure.identity import DefaultAzureCredential
from azure.identity import ManagedIdentityCredential 
from azure.mgmt.compute import ComputeManagementClient
from azure.mgmt.compute.models import DataDisk, VirtualMachine

logger = logging.getLogger(__name__)


class UpgradeCluster(Procedure):
    def __init__(self, name: str, config: ConfigParser, check_file: Path, source_volume_id: str, region: str, cloud_provider: str,
                 source_version: str, upgrade_version: str, managed_docker: bool, default_registry_url: str, python_image_name: str, mgmd_ip: str) -> None:
        self.name = name
        self.config = config
        self.check_file = check_file
        self.source_volume_id = source_volume_id
        self.region = region
        self.cloud_provider = cloud_provider
        self.source_version = HopsPlanUtils.clean_hopsworks_version(source_version)
        self.upgrade_version_uncleaned = upgrade_version
        self.upgrade_version = HopsPlanUtils.clean_hopsworks_version(upgrade_version)
        self.managed_docker = managed_docker
        self.default_registry_url = default_registry_url
        self.python_image_name = python_image_name
        self.base_image_name = "base"
        self.retries = 10
        self.waitTime = 20
        self.install_dir = "/srv/hops"
        self.hopsworks_versions = self.config.get('upgradeplan', 'hopsworks_versions').split(',')
        self.base_docker_images = self.config.get('upgradeplan', 'base_images', fallback='/root/base.tar')
        self.onlinefs_docker_image = self.config.get('upgradeplan', 'onlinefs_image', fallback='/root/onlinefs.tgz')
        self.backup_mount_dir = "/backup"
        self.backup_install_dir = f"{self.backup_mount_dir}{self.install_dir}"
        self.stage = 'Cluster upgrade. starting...'
        self.mgmd_ip = mgmd_ip

    def execute(self) -> List[Procedure]:
        try:
            # Init upgrade
            self.init_upgrade()

            # Mount backup 
            self._mount_volume()

            # Import database 
            self.import_database()

            # Import certs
            self.import_certs()

            # Import Elasticsearch 
            self.import_elastic()
           
            # Import Airflow dags
            self.import_airflow()           

            # Import Kafka data 
            self.import_kafka()

            # Import Prometheus dashboards and alerts
            self.import_prometheus()

            # Clean up HopsFS files
            self.import_hopsfs()

            # Import Flyway
            self.import_flyway()

            # Import docker images
            self.import_docker()

            # Import SSH
            self.import_ssh()

            # Restore OnlineFS API token
            self.import_onlinefs_token()

            # Every thing is OK.
            if self.check_file is not None:
                self.check_file.touch()
        except Exception as ex:
            self.friendly_name = f"{self.stage}\nError details: {ex}"
            logger.error("%s - Failed at %s", self.name, self.friendly_name)
            raise ex
        finally:
            # Umount backup
            self._umount_volume()
        return []

    def init_upgrade(self) -> None:
        self.updateHWStage("Cluster upgrade. Acquiring instance metadata")
        self.instance_id = get_cloud_provider(self.cloud_provider).get_node_id(self.config)
        logger.info("InstanceId: %s", self.instance_id)
        logger.info("Region is %s", self.region)
            

    def _mount_volume(self) -> None:
        self.updateHWStage("Cluster upgrade. Mounting old cluster")
        os.mkdir(self.backup_mount_dir)
        self.mount_volume()

    def _umount_volume(self) -> None:
        self.updateHWStage("Cluster upgrade. Unmounting old cluster")
        self.umount_volume()
        ShellUtils.rmtree(self.backup_mount_dir)
        logger.info(f"Deleted {self.backup_mount_dir}")

    @abc.abstractmethod
    def mount_volume(self) -> None:
        pass

    @abc.abstractmethod
    def umount_volume(self) -> None:
        pass

    def import_database(self) -> None:
        if self.mgmd_ip != "":
            logger.info("It is a Cluster with dedicated RonDB. Skip importing database during upgrade")
            return
        start = time.time()
        self.updateHWStage("Cluster upgrade. Importing database")
        mysql_cluster_dir = f"{self.install_dir}/mysql-cluster"
        backup_mysql_cluster_dir = f"{self.backup_mount_dir}{mysql_cluster_dir}"

        # start old cluster and backup its data using mysqldump and native ndb backups
        # copy mysql from the old cluster
        mysql_dir = f"{self.install_dir}/mysql"
        new_mysql_dir = os.readlink(mysql_dir)
        bkp_mysql_dir = os.readlink(f"{self.backup_mount_dir}{mysql_dir}")
        # copy mysql only if we are upgrading from a different mysql version
        if bkp_mysql_dir != new_mysql_dir:
            os.remove(mysql_dir)
            SystemUtils.execute_command(['cp', '-r', '-p', f"{self.backup_mount_dir}{bkp_mysql_dir}", bkp_mysql_dir], timeout=1800)
            self.chown("mysql", "mysql", bkp_mysql_dir)
            os.symlink(bkp_mysql_dir, mysql_dir)

        # copy mysql-cluster from the old cluster
        new_mysql_cluster_dir = f"{mysql_cluster_dir}-new"
        SystemUtils.execute_command(['mv',  mysql_cluster_dir, new_mysql_cluster_dir], timeout=1800)
        pcp = ParallelCpTask(self.name, backup_mysql_cluster_dir, mysql_cluster_dir, max_workers=5, timeout=3600)
        pcp.execute()
        pcp.wait_to_finish()

        self.chown("mysql", "mysql", mysql_cluster_dir)
        
        # get the old config ip and replace it with our new ip to be able to start the cluster
        bkp_config = SystemUtils.read_config(f"{mysql_cluster_dir}/config.ini", strict=False) 
        bkp_mgm_ip = bkp_config["NDB_MGMD"]["HostName"]
        local_ip = get_cloud_provider(self.cloud_provider).get_private_ip(self.config)
        
        # replace old_ip with our current ip
        sed_cmd = f"sed -i s/'{bkp_mgm_ip}'/'{local_ip}'/g"
        SystemUtils.execute_command([f"{sed_cmd} {mysql_cluster_dir}/config.ini"], shell=True, timeout=60)
        SystemUtils.execute_command([f"{sed_cmd} {mysql_cluster_dir}/my.cnf"], shell=True, timeout=60)
        SystemUtils.execute_command(["find " + mysql_cluster_dir + "/ndb/scripts -name *.sh -exec " + sed_cmd + " {} \;"], shell=True, timeout=60)
        SystemUtils.execute_command([f'echo "NDBD_INITIAL_RESTART=false" > {mysql_cluster_dir}/ndb/scripts/ndbd_env_variables'], shell=True, timeout=120)

        SystemUtils.execute_command(["systemctl", "start", "ndb_mgmd"], timeout=120)
        SystemUtils.execute_command(["systemctl", "start", "ndbmtd"], timeout=900)
        SystemUtils.execute_command(["systemctl", "start", "mysqld"], timeout=900)

        # backup databases 
        SystemUtils.execute_mysql_statement("ALTER TABLE flyway_schema_history engine = 'ndb';", "hopsworks")
        
        db_backup_path = "/root/upgrade_db_backup"
        mysql_db_dumps_path = f"{db_backup_path}/mysqldump"
        os.makedirs(mysql_db_dumps_path)

        system_databases = ["mysql", "information_schema", "performance_schema", "sys", "ndbinfo"]
        databases = [d[0] for d in SystemUtils.execute_mysql_statement("SHOW DATABASES;", "") if d[0] not in system_databases]
        for db in databases:
            SystemUtils.execute_command([f"/srv/hops/mysql/bin/mysqldump --host 127.0.0.1 --no-data --triggers --routines --events {db} > {mysql_db_dumps_path}/{db}.sql"], shell=True, timeout=900)
        
        # backup users
        SystemUtils.execute_command([f"/srv/hops/mysql/bin/mysqlpump -h 127.0.0.1 --exclude-databases=%% --exclude-users=root,mysql.sys,mysql.session,mysql.infoschema --users > {db_backup_path}/users.sql"], shell=True, timeout=900)

        # ndb native backup
        native_backup_dir = f"{db_backup_path}/ndb"
        os.makedirs(native_backup_dir)
        SystemUtils.execute_command([f"/srv/hops/mysql-cluster/ndb/scripts/mgm-client.sh -e \"START BACKUP 1 SNAPSHOTEND WAIT COMPLETED\""], shell=True, timeout=1800)
        SystemUtils.execute_command(['cp', '-r', "/srv/hops/mysql-cluster/ndb/backups/BACKUP/BACKUP-1", f"{native_backup_dir}/BACKUP-1"], timeout=1800)

        # stop the old cluster
        SystemUtils.execute_command(["systemctl", "stop", "ndb_mgmd"], timeout=120)
        SystemUtils.execute_command(["systemctl", "stop", "ndbmtd"], timeout=900)
        SystemUtils.execute_command(["systemctl", "stop", "mysqld"], timeout=900)

        # recover the new cluster
        ShellUtils.rmtree(mysql_cluster_dir)
        SystemUtils.execute_command(['mv',  new_mysql_cluster_dir, mysql_cluster_dir], timeout=1800)
        # we copy mysql only if we are upgrading from a different mysql version
        if bkp_mysql_dir != new_mysql_dir:
            os.remove(mysql_dir)
            ShellUtils.rmtree(bkp_mysql_dir)
            os.symlink(new_mysql_dir, mysql_dir)
        # enable initial start next time we start the database to wipe the database before restore
        SystemUtils.execute_command([f'echo "NDBD_INITIAL_RESTART=true" > {mysql_cluster_dir}/ndb/scripts/ndbd_env_variables'], shell=True, timeout=120)
        
        # update config.ini to create the log group and table space at the start
        # right now we assume that only 1 table space and 1 log group is needed
        # we need to do that since mysqldumps would fails due to missing table spaces at initial start
        # TODO create a table space with the same or bigger size than what was actually configured
        config_ini_file = f"{mysql_cluster_dir}/config.ini"
        config_ini_config = SystemUtils.read_config(config_ini_file, strict=False)
        initial_log_group = 'InitialLogFileGroup' in config_ini_config["NDBD DEFAULT"]
        initial_table_space = 'InitialTablespace' in config_ini_config["NDBD DEFAULT"]

        if not initial_log_group or not initial_table_space:
            config_ini_lines: List[str] = []
            with open(config_ini_file, "r") as file:
                config_ini_lines = file.readlines()

            ndbd_default_section = False
            config_ini_lines_updated: List[str] = []
            for l in config_ini_lines:                
                if ndbd_default_section and l.strip().startswith("[") and l.strip().endswith("]"):
                    if not initial_log_group:
                        config_ini_lines_updated.append("# Move this to another drive if you have a high number of ops/sec\n")
                        config_ini_lines_updated.append("InitialLogFileGroup=name=lg_1;undo_buffer_size=128M;undo_log_0.log:4G\n\n")
                    if not initial_table_space:
                        config_ini_lines_updated.append("# Move this to another drive to store small files in HopsFS\n")
                        config_ini_lines_updated.append("InitialTablespace=name=ts_1;extent_size=16M;ts_1_data_file_0.dat:8G\n\n")
                    ndbd_default_section = False

                if '[NDBD DEFAULT]' in l:
                    ndbd_default_section = True      
                config_ini_lines_updated.append(l)

            with open(config_ini_file, "w") as file:
                file.writelines(config_ini_lines_updated)

        logger.info("%s took %d seconds to import database", self.name, (time.time() - start))

    def import_certs(self) -> None:
        self.updateHWStage("Cluster upgrade. Importing CA certs")
        logger.info("Copying CA certs from snapshot.")
        certs_dir_name = "certs-dir"
        certs_dir = f"{self.install_dir}/{certs_dir_name}"        
        self.replace_dir(self.backup_install_dir, self.install_dir, certs_dir_name)
        # Fix certs-dir ownership explicitly (TODO implement a better solution)
        # /srv/hops/certs-dir/certs 
        self.chown("glassfish", "glassfish", f"{certs_dir}/certs")
        self.chown("root", "root", f"{certs_dir}/certs/*")

        # /srv/hops/certs-dir/crl 
        self.chown("glassfish", "glassfish", f"{certs_dir}/crl")

        # /srv/hops/certs-dir/encryption_master_password
        self.chown("glassfish", "glassfish", f"{certs_dir}/encryption_master_password")

        # /srv/hops/certs-dir/hops-site-certs
        self.chown("glassfish", "certs", f"{certs_dir}/hops-site-certs")
        self.chown("glassfish", "root", f"{certs_dir}/hops-site-certs/keystores")

        # /srv/hops/certs-dir/index.txt*
        self.chown("root", "root", f"{certs_dir}/index.txt*")

        # /srv/hops/certs-dir/intermediate
        intermediate_dir = f"{certs_dir}/intermediate"
        self.chown("glassfish", "glassfish", intermediate_dir)
        self.chown("root", "root", f"{intermediate_dir}/certs/*")
        self.chown("glassfish", "glassfish", f"{intermediate_dir}/crl/*")
        self.chown("root", "root", f"{intermediate_dir}/crlnumber*")
        self.chown("root", "root", f"{intermediate_dir}/csr/*")
        self.chown("root", "root", f"{intermediate_dir}/index.txt*")
        self.chown("root", "root", f"{intermediate_dir}/newcerts/*")
        self.chown("glassfish", "root", f"{intermediate_dir}/openssl-intermediate.cnf")
        self.chown("glassfish", "glassfish", f"{intermediate_dir}/private/*")
        self.chown("root", "root", f"{intermediate_dir}/serial*")

        # /srv/hops/certs-dir/newcerts
        self.chown("glassfish", "glassfish", f"{certs_dir}/newcerts")
        self.chown("root", "root", f"{certs_dir}/newcerts/*")

         # /srv/hops/certs-dir/openssl-ca.cnf
        self.chown("glassfish", "root", f"{certs_dir}/openssl-ca.cnf")

         # /srv/hops/certs-dir/private
        self.chown("glassfish", "glassfish", f"{certs_dir}/private")
        self.chown("root", "root", f"{certs_dir}/private/*")

        # /srv/hops/certs-dir/serial*
        self.chown("root", "root", f"{certs_dir}/serial*")

        # /srv/hops/certs-dir/transient
        transient_dir = f"{certs_dir}/transient"
        self.chown("glassfish", "userscerts", transient_dir)
        ShellUtils.rmtree(transient_dir, keep_root=True)

        # Revoke all old service certificates
        # If possible do not include consul TLD as we might need to change them in the future
        certs_to_revoke = ["L=consul", "L=elastic", "L=elkadmin", "L=flink", "L=glassfish", "L=hdfs",
                           "L=hive", "L=hopsmon", "L=kafka", "L=kagent", "L=livy", "L=mapred", "L=mysql",
                           "L=rmyarn", "L=yarn", "L=onlinefs", "CN=hopsworks.glassfish.service"]
        PlanUtils.revoke_old_certificates(self.config, certs_dir, certs_to_revoke)

    def import_elastic(self) -> None:
        # TODO use snapshot/restore APIs of elasticsearch
        self.updateHWStage("Cluster upgrade. Importing Elasticsearch Indices")
        self.replace_dir(self.backup_install_dir, self.install_dir, "elasticsearch-data")

    def import_airflow(self) -> None:
        self.updateHWStage("Cluster upgrade. Importing Airflow DAGs")
        airflow_dir = f"{self.install_dir}/airflow"
        backup_airflow_dir = f"{self.backup_mount_dir}{airflow_dir}"

        self.replace_dir(backup_airflow_dir, airflow_dir, "dags")
        airflow_dags_dir = f"{airflow_dir}/dags"
        for s_dir in os.listdir(airflow_dags_dir):
            self.chown("glassfish", "glassfish", f"{airflow_dags_dir}/{s_dir}")

    def import_kafka(self) -> None:
        self.updateHWStage("Cluster upgrade. Importing Kafka topics")
        self.replace_dir(self.backup_install_dir, self.install_dir, "kafka-logs")

    def import_prometheus(self) -> None:
        self.updateHWStage("Cluster upgrade. Importing Prometheus dashboards and alerts")
        prometheus_dir = f"{self.install_dir}/prometheus"
        backup_prometheus_dir = f"{self.backup_mount_dir}{prometheus_dir}"

        self.replace_dir(backup_prometheus_dir, prometheus_dir, "prometheus-data")

        alertmanager_dir = f"{self.install_dir}/alertmanager"
        backup_alertmanager_dir = f"{self.backup_mount_dir}{alertmanager_dir}"

        self.replace_dir(backup_alertmanager_dir, alertmanager_dir, "alertmanager-data")

    def import_hopsfs(self) -> None:
        self.updateHWStage("Cluster upgrade. Importing HopsFS files")
        logger.info("Deleting datanodes data dirs")
        ShellUtils.rmtree(f"{self.install_dir}/hopsdata/hdfs/dn/cloud/current/")
        ShellUtils.rmtree(f"{self.install_dir}/hopsdata/hdfs/dn/disk/current/")

    def import_flyway(self) -> None:
        self.updateHWStage("Cluster upgrade. Importing schema files")
        flyway_dir = f"{self.install_dir}/domains/domain1/flyway"
        backup_flyway_dir = f"{self.backup_mount_dir}{flyway_dir}"

        self.replace_dir(backup_flyway_dir, flyway_dir, "sql")
        self.replace_dir(backup_flyway_dir, flyway_dir, "undo")

        start_flyway_migration = False
        for version in self.hopsworks_versions:
            if self.source_version == version:
                start_flyway_migration = True
                continue
            if start_flyway_migration:
                shutil.copy2(f"{flyway_dir}/all/sql/V{version}__hopsworks.sql", f"{flyway_dir}/sql")
                shutil.copy2(f"{flyway_dir}/all/undo/U{version}__undo.sql", f"{flyway_dir}/undo")
                
            if self.upgrade_version == version:
                break
    
    def import_ssh(self) -> None:
        self.updateHWStage("Cluster upgrade. Restoring ssh permissions")
        if os.path.isdir("/home/ubuntu/.ssh") and os.path.isdir(f"{self.backup_mount_dir}/home/ubuntu/.ssh"):
            logger.info("Copying .ssh folder for Ubuntu user")
            self.replace_dir(f"{self.backup_mount_dir}/home/ubuntu", "/home/ubuntu", ".ssh")
        
    def import_docker(self) -> None:
        if self.managed_docker:
            return
        start = time.time()
        self.updateHWStage("Cluster upgrade. Import docker images")
        self.retry_operation_on_exception(lambda: SystemUtils.execute_command(["systemctl", "stop", "docker"], timeout=600))
        

        #TODO we need to change to a more efficient backup mechanism using docker load/save
        docker_dir = f"{self.install_dir}/docker"
        backup_docker_dir = f"{self.backup_mount_dir}{docker_dir}"
        pcp = ParallelCp(ParallelCp.__name__, backup_docker_dir, docker_dir, max_workers=5, timeout=3600, overwrite=True)
        pcp.execute()

        self.retry_operation_on_exception(lambda: SystemUtils.execute_command(["systemctl", "start", "docker"], timeout=600))

        if os.path.exists(self.base_docker_images):
            docker_client = docker.from_env(timeout=3600)
            with open(self.base_docker_images, 'rb') as f:
                docker_client.images.load(f)
            base_image_tag = f"{self.base_image_name}:{self.upgrade_version_uncleaned}"
            base_image = docker_client.images.get(base_image_tag)
            base_image.tag(f"{self.default_registry_url}/{self.base_image_name}:{self.upgrade_version_uncleaned}")
            docker_client.images.remove(base_image_tag)

            python_image_tag = f"{self.python_image_name}:{self.upgrade_version_uncleaned}"
            python_image = docker_client.images.get(python_image_tag)
            python_image.tag(f"{self.default_registry_url}/{self.python_image_name}:{self.upgrade_version_uncleaned}")
            docker_client.images.remove(python_image_tag)
        else:
            logger.warning("%s base images doesn't exists locally %s", self.name, self.base_docker_images)
                    
        logger.info("%s took %d seconds to import docker", self.name, (time.time() - start))
 
    def import_onlinefs_token(self) -> None:
        self.updateHWStage("Cluster upgrade. Restoring OnlineFS API token")
        token_file = f"{self.install_dir}/onlinefs/etc/token"
        backup_token_file = f"{self.backup_mount_dir}{token_file}"
        if os.path.exists(backup_token_file):
            logger.info(f"Restoring OnlineFS API token to {token_file}")
            self.replace_file(backup_token_file, token_file)
        else:
            logger.debug(f"Skip restoring OnlineFS API because {backup_token_file} does not exist")

        if os.path.exists(self.onlinefs_docker_image):
            docker_client = docker.from_env(timeout=3600)
            with open(self.onlinefs_docker_image, 'rb') as f:
                docker_client.images.load(f)

    def rollback(self) -> None:
        pass
    
    def retry_operation_on_exception(self, op_function: Callable) -> Any:
        for retry in range(self.retries):
            try:
                op_resp = op_function()
            except Exception as ex:
                logger.warning("Operation failed. Retries %d", retry)
                exception = ex
                time.sleep(self.waitTime)
                continue
            return op_resp
        raise exception
    
    def replace_dir(self, src: str,  dst: str, dir_name: str) -> None:
        src_path = f"{src}/{dir_name}"
        dst_path = f"{dst}/{dir_name}"
        owner = Path(dst_path).owner()
        group = Path(dst_path).group()
        if len(os.listdir(src_path)) > 0:
            ShellUtils.rmtree(dst_path, keep_root=True)
            SystemUtils.execute_command([f"cp -r -p {src_path}/* {dst_path}"], shell=True, timeout=1800)
            self.chown(owner, group, f"{dst_path}/*")

    def replace_file(self, src: str,  dst: str) -> None:
        shutil.copy2(src, dst)

    def chown(self, owner: str, group: str, path : str) -> None:
        SystemUtils.execute_command([f"chown -R {owner}:{group} {path}"], shell=True, timeout=1800)

    def mount(self, device: str, folder: str) -> None:
        for retry in range(self.retries):
            if os.path.exists(device):
                try:
                    SystemUtils.execute_command(['mount', '-o', 'ro', device, folder], timeout=300)
                except Exception as ex:
                    raise ex
                return
            else:
                logger.info("The device %s does not exist yet. Retry count: %d", device, retry)
                time.sleep(self.waitTime)
                continue
        raise Exception(f"Failed to mount volume {device}")

    def umount(self,  folder: str) -> None:
        SystemUtils.execute_command(['umount', folder], timeout=300)

    def updateHWStage(self, stage: str) -> None:
        self.stage = stage
        if self.report_callback is not None:
            try:
                self.report_callback.set_stage(stage) # type: ignore
                self.report_callback.execute()
            except Exception as ex:
                logger.warning("%s - Could not execute pre-execute: %s", self.name, ex)


class AWSUpgradeCluster(UpgradeCluster):
    def __init__(self, name: str, config: ConfigParser, check_file: Path, source_volume_id: str, region: str,
                 cloud_provider: str, source_version: str, upgrade_version: str, managed_docker: bool, default_registry_url: str,
                 python_image_name: str, mgmd_ip: str) -> None:
      super().__init__(name, config, check_file, source_volume_id, region, cloud_provider,
                       source_version, upgrade_version, managed_docker, default_registry_url, python_image_name, mgmd_ip)
      self.ec2: Client = boto3.client('ec2', region_name=self.region)

    def mount_volume(self) -> None:
        response = self.ec2.describe_volumes(VolumeIds=[self.source_volume_id])
        logger.info("Describe volume  %s - response : %s", self.source_volume_id, str(response))
        if len(response['Volumes']) != 1 and len(response['Volumes'][0]['Attachments']) != 1:
            raise Exception("Something went wrong. Could not find the disk attached to the head node!")

        source_attachment = response['Volumes'][0]['Attachments'][0]
        self.source_instance_id = source_attachment['InstanceId']
        self.source_device_name = source_attachment['Device']
        self.source_delete_on_termination = source_attachment['DeleteOnTermination']

        self._detach_volume(self.source_volume_id, self.source_instance_id)
        self._attach_volume(self.source_volume_id, self.instance_id)
        
        # lsblk -o +SERIAL | grep source_volume_id
        # https://docs.aws.amazon.com/AWSEC2/latest/UserGuide/nvme-ebs-volumes.html 
        # Typically /dev/nvme1n1p1
        deviceNameOnInstance = f"/dev/disk/by-id/nvme-Amazon_Elastic_Block_Store_{self.source_volume_id.replace('-', '')}-part1"
        self.mount(deviceNameOnInstance, self.backup_mount_dir)
        logger.info("Mounted volume %s on dir %s", self.source_volume_id, self.backup_mount_dir)
    
    def umount_volume(self) -> None:
        self.umount(self.backup_mount_dir)
        logger.info("Unmounted Volume %s", self.source_volume_id)
        self._detach_volume(self.source_volume_id, self.instance_id)
        self._attach_volume(self.source_volume_id, self.source_instance_id, self.source_device_name)
        self._modify_volume_delete_on_termination(self.source_volume_id, self.source_instance_id, self.source_device_name, self.source_delete_on_termination)
    
    def _detach_volume(self, volume_id: str, instance_id: str) -> None:
        logger.info("Detaching volume %s from instance %s", volume_id, instance_id)
        self.retry_operation_on_exception(lambda: self.ec2.detach_volume(
            VolumeId=volume_id,
            InstanceId=instance_id,
        ))

    def _attach_volume(self, volume_id: str, instance_id: str, device: str = '/dev/sdm') -> None:
        logger.info("Attaching volume %s to instance %s with device name %s", volume_id, instance_id, device)
        self.retry_operation_on_exception(lambda: self.ec2.attach_volume(
            VolumeId=volume_id,
            InstanceId=instance_id,
            Device=device
        ))

    def _modify_volume_delete_on_termination(self, volume_id: str, instance_id: str, device_name: str, delete_on_termination: bool) -> None:
        logger.info("Modifying volume %s on instance %s with device name %s to delete_on_termination %s", volume_id, instance_id, device_name, str(delete_on_termination))
        self.retry_operation_on_exception(lambda: self.ec2.modify_instance_attribute(
            InstanceId=instance_id,
            BlockDeviceMappings=[{
                'DeviceName': device_name,
                'Ebs' : {
                    'DeleteOnTermination' : delete_on_termination,
                    'VolumeId': volume_id
                }
            }]
        ))

class AzureUpgradeCluster(UpgradeCluster):
    def __init__(self, name: str, config: ConfigParser, check_file: Path, source_volume_id: str, region: str,
                 cloud_provider: str, source_version: str, upgrade_version: str, managed_docker: bool, default_registry_url: str,
                 python_image_name: str, managed_identity_client_id: str, resource_group: str, subscription_id: str, mgmd_ip: str) -> None:
      super().__init__(name, config, check_file, source_volume_id, region, cloud_provider,
                       source_version, upgrade_version, managed_docker, default_registry_url, python_image_name, mgmd_ip)
      self.managed_identity_client_id = managed_identity_client_id
      self.resource_group = resource_group
      self.subscription_id = subscription_id
      if self.managed_identity_client_id:
          logger.info("Using managed identity credentials with client ID: %s", self.managed_identity_client_id)
          credential = ManagedIdentityCredential(client_id=self.managed_identity_client_id)
      else:
          credential = DefaultAzureCredential()
      self.compute_client = ComputeManagementClient(
          credential, self.subscription_id)

    def mount_volume(self) -> None:
        disk = self.retry_operation_on_exception(lambda: self.compute_client.disks.get(self.resource_group, self.source_volume_id))
        disk_lun = self._attach_disk(disk.id, disk.name)
        # https://docs.microsoft.com/en-us/azure/virtual-machines/troubleshooting/troubleshoot-device-names-problems
        # typically /dev/sdc1
        dev_disk = f"/dev/disk/azure/scsi1/lun{disk_lun}-part1"
        self.mount(dev_disk, self.backup_mount_dir)
        logger.info("Mounted disk %s on dir %s", self.source_volume_id, self.backup_mount_dir)

    def umount_volume(self) -> None:
        self.umount(self.backup_mount_dir)
        self._detach_disk(self.source_volume_id)
        logger.info("Unmounted disk %s", self.source_volume_id)
    
    def _attach_disk(self, disk_id: str, disk_name: str) -> int :
        vm_info : VirtualMachine = self.retry_operation_on_exception(lambda: self.compute_client.virtual_machines.get(
            self.resource_group, self.instance_id))
        attached_disk_lun_nums = [d.lun for d in vm_info.storage_profile.data_disks]
        disk_lun = 0
        while disk_lun in attached_disk_lun_nums:
            disk_lun = disk_lun + 1
        
        logger.info("Attaching disk %s to instance %s with lun %d", disk_name, self.instance_id, disk_lun)

        updated_data_disks = vm_info.storage_profile.data_disks + [{
                "name": disk_name,
                "lun": disk_lun,
                "create_option": "Attach",
                "managed_disk": {
                    "id": disk_id
                }
            }]

        vm_attach_disk = {
            "storage_profile": {
                "data_disks": updated_data_disks
            }
        }
        poller = self.retry_operation_on_exception(lambda: self.compute_client.virtual_machines.begin_update(
            self.resource_group, self.instance_id, vm_attach_disk))
        poller.result()
        return disk_lun

    def _detach_disk(self, disk_name: str) -> None:
        logger.info("Detaching disk %s from instance %s", disk_name, self.instance_id)
        vm_info : VirtualMachine = self.retry_operation_on_exception(lambda: self.compute_client.virtual_machines.get(
            self.resource_group, self.instance_id))
        updated_data_disks = [d for d in vm_info.storage_profile.data_disks if d.name != disk_name]
        vm_detach_disk = {
            "storage_profile": {
                "data_disks": updated_data_disks
            }
        }
        poller = self.retry_operation_on_exception(lambda: self.compute_client.virtual_machines.begin_update(
            self.resource_group, self.instance_id, vm_detach_disk))
        poller.result()
