import base64
from ec2init.plans.preinit import grow_data_volume
import logging
import tempfile
from configparser import ConfigParser
import json
import pathlib
import abc

from typing import Sequence, List

from ec2init import Procedure, Plan, plans
from ec2init.plans.init.callback_hopsworks import CallbackHopsworks
from ec2init.plans.upgrade.upgrade_cluster import AWSUpgradeCluster, AzureUpgradeCluster
from ec2init.plans.init.base_init_plan import MasterBaseInitPlan

from ec2init.utils.system import SystemUtils
from ec2init.utils.http import HttpUtils
from ec2init.plans.preinit.grow_data_volume import GrowDataVolume

logger = logging.getLogger(__name__)


class UpgradePlan(MasterBaseInitPlan):
    def init(self, name: str, config: ConfigParser, **kwargs: str) -> None:
        super().init(name, config, **kwargs)

    def create(self) -> Sequence[Procedure]:
        plan: List[Procedure] = []

        # Make a callback to Hopsworks Cloud indicating the instance has booted
        self.add_callback_instance_booted(plan)

        # Grow data volume in multi-disk setup
        self.add_grow_data_volume(plan)

        # Upgrading Instance
        self.add_upgrade_cluster(plan)

        return plan

    @abc.abstractmethod
    def add_upgrade_cluster(self, plan: List[Procedure]):
        pass


class AWSUpgradePlan(UpgradePlan):

    def add_upgrade_cluster(self, plan: List[Procedure]):
        if self.source_volume_id:
            grow_data_volume_check_file = self.plan_utils.get_procedure_check_file(GrowDataVolume.__name__)
            upgrade_check_file = self.plan_utils.get_procedure_check_file(AWSUpgradeCluster.__name__)
            if not self.plan_utils.check_file_exists(upgrade_check_file)\
                 and self.plan_utils.check_file_exists(grow_data_volume_check_file):
                logger.info("%s - Adding procedure %s", self.name, AWSUpgradeCluster.__name__)
                logger.info("%s - Upgrading from %s to %s", self.name, self.source_version, self.version)
                upgradeplan = AWSUpgradeCluster(AWSUpgradeCluster.__name__, self.config, check_file=upgrade_check_file, source_volume_id=self.source_volume_id,
                                                region=self.region, cloud_provider=self.cloud_provider, source_version=self.source_version, 
                                                upgrade_version=self.version, managed_docker=bool(self.user_registry_url and self.main_registry_url),
                                                default_registry_url=self.default_registry_url, python_image_name=self.python_image_name, mgmd_ip=self.mgmd_ip)
                upgradeplan.friendly_name = "Cluster upgrade. Restoring data from old cluster "
                upgradeplan.fail_plan = True
                upgradeplan.report_callback = self.plan_utils.create_callback_procedure("starting", stage="Cluster Upgrade",
                                                                                        instance_type=self.instance_type)
                plan.append(upgradeplan)
        else:
            logger.info("Skipping upgrade. No volume ID specified")


class AzureUpgradePlan(UpgradePlan):
    
    def init(self, name: str, config: ConfigParser, **kwargs: str) -> None:
        super().init(name, config, **kwargs)
        self.azure_managed_identity_id = kwargs.get("azure_managed_identity_id", "")
        self.azure_resource_group = kwargs.get("azure_resource_group", "")
        self.azure_subscription_id = kwargs.get("azure_subscription_id", "")

    def add_upgrade_cluster(self, plan: List[Procedure]):
        if self.source_volume_id:
            grow_data_volume_check_file = self.plan_utils.get_procedure_check_file(GrowDataVolume.__name__)
            upgrade_check_file = self.plan_utils.get_procedure_check_file(AzureUpgradeCluster.__name__)
            if not self.plan_utils.check_file_exists(upgrade_check_file)\
                and self.plan_utils.check_file_exists(grow_data_volume_check_file):
                logger.info("%s - Adding procedure %s", self.name, AzureUpgradeCluster.__name__)
                logger.info("%s - Upgrading from %s to %s", self.name, self.source_version, self.version)
                upgradeplan = AzureUpgradeCluster(AzureUpgradeCluster.__name__, self.config, check_file=upgrade_check_file, source_volume_id=self.source_volume_id,
                                                region=self.region, cloud_provider=self.cloud_provider, source_version=self.source_version, 
                                                upgrade_version=self.version, managed_docker=bool(self.user_registry_url and self.main_registry_url),
                                                default_registry_url=self.default_registry_url, python_image_name=self.python_image_name,
                                                managed_identity_client_id=self.azure_managed_identity_id, resource_group=self.azure_resource_group, subscription_id=self.azure_subscription_id,
                                                mgmd_ip=self.mgmd_ip)
                upgradeplan.friendly_name = "Cluster upgrade. Restoring data from old cluster "
                upgradeplan.fail_plan = True
                upgradeplan.report_callback = self.plan_utils.create_callback_procedure("starting", stage="Cluster Upgrade",
                                                                                        instance_type=self.instance_type)
                plan.append(upgradeplan)
        else:
            logger.info("Skipping upgrade. No volume ID specified")