import logging
import time
import re

from .system import SystemUtils
from .http import HttpUtils
from configparser import ConfigParser

from typing import Pattern, Type
from typing import Dict

logger = logging.getLogger(__name__)

_ec2_meta_server = "http://169.254.169.254"
_ec2_metadata_version = "2016-09-02"

_AWS_PUBLIC_IP_ENDPOINT = (
    "GET", f"{_ec2_meta_server}/{_ec2_metadata_version}/meta-data/public-ipv4")
_AWS_LOCAL_IP_ENDPOINT = (
    "GET", f"{_ec2_meta_server}/{_ec2_metadata_version}/meta-data/local-ipv4")
_AWS_FQDN_ENDPOINT = (
    "GET", f"{_ec2_meta_server}/{_ec2_metadata_version}/meta-data/local-hostname")
_AWS_INSTANCE_ID_ENDPOINT = (
    "GET", f"{_ec2_meta_server}/{_ec2_metadata_version}/meta-data/instance-id")
_AWS_AVAILABILITY_ZONE = (
    "GET", f"{_ec2_meta_server}/{_ec2_metadata_version}/meta-data/placement/availability-zone")


_azure_meta_server = "http://169.254.169.254"
_azure_metadata_version = "2019-11-01"
_azure_http_headers = {"Metadata": "true"}

_AZURE_INSTANCE_ID_ENDPOINT = (
    "GET", f"{_azure_meta_server}/metadata/instance/compute/name?format=text&api-version={_azure_metadata_version}")
_AZURE_PUBLIC_IP_ENDPOINT = (
    "GET", f"{_azure_meta_server}/metadata/instance/network/interface/0/ipv4/ipAddress/0/publicIpAddress?format=text&api-version={_azure_metadata_version}")
_AZURE_PRIVATE_IP_ENDPOINT = (
    "GET", f"{_azure_meta_server}/metadata/instance/network/interface/0/ipv4/ipAddress/0/privateIpAddress?format=text&api-version={_azure_metadata_version}")
_AZURE_AVAILABILITY_ZONE = (
    "GET", f"{_azure_meta_server}/metadata/instance/compute/zone?format=text&api-version={_azure_metadata_version}")

_DEFAULT_AZURE_SEARCH_DOMAIN = ".internal.cloudapp.net"

_gcp_meta_server = "http://metadata.google.internal"
_GCP_FQDN_ENDPOINT = ("GET", f"{_gcp_meta_server}/computeMetadata/v1/instance/hostname")
_GCP_PUBLIC_IP_ENDPOINT = ("GET", f"{_gcp_meta_server}/computeMetadata/v1/instance/network-interfaces/0/access-configs/0/external-ip")
_GCP_PRIVATE_IP_ENDPOINT = ("GET", f"{_gcp_meta_server}/computeMetadata/v1/instance/network-interfaces/0/ip")
_GCP_INSTANCE_ID_ENDPOINT = ("GET", f"{_gcp_meta_server}/computeMetadata/v1/instance/name")
_GCP_AVAILABILITY_ZONE = ("GET", f"{_gcp_meta_server}/computeMetadata/v1/instance/zone")

class CloudProvider:
    @staticmethod
    def get_fqdn(config: ConfigParser, search_domain: str, node_id: str) -> str:
        pass

    @staticmethod
    def get_hosted_zone(config: ConfigParser) -> str:
        hosted_zone = config.get('initplan', 'hosted_zone', fallback='cloud.hopsworks.ai')
        if SystemUtils.is_production_environment():
            return hosted_zone
        tokens = hosted_zone.split(".")
        tokens[0] = "dev-cloud"
        return ".".join(tokens)

    @staticmethod
    def get_public_domain_name(config: ConfigParser, cluster_id: str, cluster_domain_prefix: str) -> str:
        hosted_zone = CloudProvider.get_hosted_zone(config)
        name = cluster_id
        if cluster_domain_prefix!="":
            name = cluster_domain_prefix
        if hosted_zone.startswith('.'):
            return f"{name}{hosted_zone}"
        return f"{name}.{hosted_zone}"

    @staticmethod
    def get_public_ip(config: ConfigParser) -> str:
        pass

    @staticmethod
    def get_private_ip(config: ConfigParser) -> str:
        pass

    @staticmethod
    def get_node_id(config: ConfigParser) -> str:
        pass

    @staticmethod
    def get_availability_zone(config: ConfigParser) -> str:
        pass

    @staticmethod
    def select_dns_name(fqdn: str, search_domain: str) -> str:
        pass


def _call_metadata_server(config: ConfigParser, endpoint: tuple, headers: Dict[str, str] = {}) -> str:
    method, url = endpoint
    response = HttpUtils.do_api_call(
        config, url, method, headers=headers)
    return response['response']

def get_azure_search_domain(search_domain: str) -> str:
    if search_domain.startswith("."):
        return search_domain
    else :
        return "." + search_domain

class Azure(CloudProvider):
    @staticmethod
    def get_fqdn(config: ConfigParser, search_domain: str, node_id: str) -> str:
        try:
            fqdn = f"{node_id}{get_azure_search_domain(search_domain)}"
            #check that the provided search_domain and host_name give a resolvable fqdn
            ip = SystemUtils.execute_command(["dig", fqdn, "+short"]).stdout.strip().decode("utf-8").replace('"','')
            logger.debug(f"fqdn is {fqdn} and ip is {ip} while private ip is {Azure.get_private_ip(config)}")
            private_ip = Azure.get_private_ip(config)
        except Exception as ex:
            raise Exception("Cluster initialization failed, could not get local hostname")
        if ip != private_ip:
            raise Exception(f"the provided host name ({node_id}) and search domain ({search_domain}) do not form a resolvable fqdn")
        return fqdn

    @staticmethod
    def get_public_ip(config: ConfigParser) -> str:
        return _call_metadata_server(config, _AZURE_PUBLIC_IP_ENDPOINT, _azure_http_headers)

    @staticmethod
    def get_private_ip(config: ConfigParser) -> str:
        return _call_metadata_server(config, _AZURE_PRIVATE_IP_ENDPOINT, _azure_http_headers)

    @staticmethod
    def get_node_id(config: ConfigParser) -> str:
        return _call_metadata_server(config, _AZURE_INSTANCE_ID_ENDPOINT, _azure_http_headers)

    @staticmethod
    def get_availability_zone(config: ConfigParser) -> str:
        return _call_metadata_server(config, _AZURE_AVAILABILITY_ZONE, _azure_http_headers)

    @staticmethod
    def select_dns_name(fqdn: str, search_domain: str) -> str:
        if "\n" not in fqdn:
            return fqdn
        candidates = list(filter(lambda x: get_azure_search_domain(search_domain) in x, fqdn.split("\n")))
        if len(candidates) < 1:
            return Azure.__remove_trailing_dot(fqdn.split("\n")[0])
        return Azure.__remove_trailing_dot(candidates[0])

    @staticmethod
    def __remove_trailing_dot(fqdn: str) -> str:
        if fqdn.endswith("."):
            return fqdn[:-1]
        return fqdn



class AWS(CloudProvider):
    @staticmethod
    def get_fqdn(config: ConfigParser, search_domain: str, node_id: str) -> str:
        try:
            return _call_metadata_server(config, _AWS_FQDN_ENDPOINT)
        except Exception as ex:
            raise Exception("Cluster initialization failed, could not get local hostname")

    @staticmethod
    def get_public_ip(config: ConfigParser) -> str:
        return _call_metadata_server(config, _AWS_PUBLIC_IP_ENDPOINT)

    @staticmethod
    def get_private_ip(config: ConfigParser) -> str:
        return _call_metadata_server(config, _AWS_LOCAL_IP_ENDPOINT)

    @staticmethod
    def get_node_id(config: ConfigParser) -> str:
        return _call_metadata_server(config, _AWS_INSTANCE_ID_ENDPOINT)
    
    @staticmethod
    def get_availability_zone(config: ConfigParser) -> str:
        return _call_metadata_server(config, _AWS_AVAILABILITY_ZONE)

    @staticmethod
    def select_dns_name(fqdn: str, search_domain: str) -> str:
        return fqdn


class GCP(CloudProvider):
    @staticmethod
    def get_fqdn(config: ConfigParser, search_domain: str, node_id: str) -> str:
        try:
            return _call_metadata_server(config, _GCP_FQDN_ENDPOINT, { "Metadata-Flavor": "Google" })
        except Exception as ex:
            raise Exception("Cluster initialization failed, could not get local hostname")

    @staticmethod
    def get_public_ip(config: ConfigParser) -> str:
        return _call_metadata_server(config, _GCP_PUBLIC_IP_ENDPOINT, { "Metadata-Flavor": "Google" })

    @staticmethod
    def get_private_ip(config: ConfigParser) -> str:
        return _call_metadata_server(config, _GCP_PRIVATE_IP_ENDPOINT, { "Metadata-Flavor": "Google" })

    @staticmethod
    def get_node_id(config: ConfigParser) -> str:
        return _call_metadata_server(config, _GCP_INSTANCE_ID_ENDPOINT, { "Metadata-Flavor": "Google" })
    
    @staticmethod
    def get_availability_zone(config: ConfigParser) -> str:
        return _call_metadata_server(config, _GCP_AVAILABILITY_ZONE, { "Metadata-Flavor": "Google" })

    @staticmethod
    def select_dns_name(fqdn: str, search_domain: str) -> str:
        return fqdn


class Provider():
    AZURE = 'azure'
    AWS = 'aws'
    GCP = 'gcp'


_cloud_provider_factory = {
    Provider.AZURE: Azure,
    Provider.AWS: AWS,
    Provider.GCP: GCP
}


def get_cloud_provider(provider: str) -> Type[CloudProvider]:
    # Fallback on AWS to be backwards compatible in places where cloud_provider is not set
    return _cloud_provider_factory.get(provider, AWS)
