import configparser
import subprocess
import logging
import os
import multiprocessing
from enum import Enum
from pathlib import Path

from typing import List, Dict, Optional, Type

logger = logging.getLogger(__name__)

MYSQL_CLIENT = "/srv/hops/mysql-cluster/ndb/scripts/mysql-client.sh"

class OSType(Enum):
    UBUNTU = 1
    CENTOS = 2

class SystemUtils:

    UBUNTU_VERSION_FILE = Path("/etc/debian_version")
    CENTOS_RELEASE_FILE = Path("/etc/centos-release")

    @staticmethod
    def execute_command(command: List[str], check: bool = True, env: Dict[str, str] = None,
                        timeout: Optional[int] = 20, shell: bool = False) -> subprocess.CompletedProcess:
        logger.debug("Executing %s", command)

        return subprocess.run(command, stdout=subprocess.PIPE, check=check,
                              env=env, timeout=timeout, shell=shell)

    @staticmethod
    def execute_mysql_statement(statement: str, database: str = "hopsworks", timeout: int = 120) -> List[List[str]]:
        logger.debug("Executing %s", statement)
        process = SystemUtils.execute_command([MYSQL_CLIENT, database, "-e", statement, "-sN"], check=True, timeout=timeout, shell=False)
        out = process.stdout.decode('latin1')
        rows = [line.strip().split("\t") for line in out.split("\n") if line and not line.startswith('Using socket')]
        return rows

    @staticmethod
    def read_config(config_file: str, strict: bool = True) -> configparser.ConfigParser:
        config = configparser.ConfigParser(strict=strict)
        config.read(config_file)
        return config

    @staticmethod
    def is_production_environment() -> bool:
        env = SystemUtils.get_environment()
        return env is not None and env.lower() == "prod"

    @staticmethod
    def is_staging() -> bool:
        stag = SystemUtils.get_staging()
        return not SystemUtils.is_production_environment() or (stag is not None and stag.lower() == "true")

    @staticmethod
    def get_environment() -> Optional[str]:
        return os.environ.get('APP_ENV')

    @staticmethod
    def get_staging() -> Optional[str]:
        return os.environ.get('CERT_STAGING')

    # Get system memory in Mbytes
    @staticmethod
    def get_system_memory() -> int:
        mem_bytes = os.sysconf('SC_PAGE_SIZE') * os.sysconf('SC_PHYS_PAGES')
        return int(mem_bytes / (1024 * 1024))

    @staticmethod
    def get_num_cpus() -> int:
        return multiprocessing.cpu_count()

    @staticmethod
    def get_os_type() -> OSType:
        if SystemUtils.CENTOS_RELEASE_FILE.exists():
            return OSType.CENTOS
        if SystemUtils.UBUNTU_VERSION_FILE.exists():
            return OSType.UBUNTU
        raise Exception('Unable to determine OS type')

    @staticmethod
    def execute_hdfs_command(command: str, timeout: int=320) -> subprocess.CompletedProcess:
        return SystemUtils.execute_command([f"su - hdfs -c '/srv/hops/hadoop/bin/hdfs {command} '" ], timeout=timeout, shell=True)